/*
 * Copyright (c) 2007, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.registry;

import org.wso2.registry.utils.AuthorizationUtils;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.utils.Transaction;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.List;

public class CollectionImpl extends ResourceImpl implements Collection {

    private static Log log = LogFactory.getLog(CollectionImpl.class);

    protected int childCount;

    public CollectionImpl() {
        childCount = -1;
    }

    public CollectionImpl(String[] paths) {
        content = paths;
        childCount = paths.length;
    }

    public void setContent(Object content) throws RegistryException {

        // note that string contents are allowed in collection to support custom generated UIs.
        if (content instanceof String[] ||
                content instanceof Resource[] ||
                content instanceof String) {
            super.setContent(content);
            return;
        }
        throw new IllegalArgumentException("Invalid content for collection. " +
                "Content of type " + content.getClass().toString() +
                " is not allowed for collections.");
    }

    public String[] getChildren() throws RegistryException {
        return (String[]) getContent();
    }

    public String[] getChildren(int start, int pageLen) throws RegistryException {

        if (content == null) {

            List <String> childPathList;

            if (Transaction.isStarted()) {

                childPathList = resourceDAO.
                        getChildPaths(getId(), start, pageLen, Transaction.getConnection());

            } else {

                Connection conn = null;

                try {
                    conn = dataSource.getConnection();
                    childPathList = resourceDAO.getChildPaths(getId(), start, pageLen, conn);

                } catch (SQLException e) {

                    String msg = "Failed to get the child paths " + pageLen + " child paths from " +
                            start + " of resource " + getPath() + ". " + e.getMessage();
                    log.error(msg, e);
                    throw new RegistryException(msg, e);

                } finally {

                    if (conn != null) {
                        try {
                            conn.close();
                        } catch (SQLException e) {
                            log.error("Failed to close the database connection used to " +
                                    "get child paths of collection " + getPath());
                        }
                    }
                }
            }

            return childPathList.toArray(new String[childPathList.size()]);


            //try {
            //    String sql = "SELECT R.RID, R.PATH  FROM REG_RESOURCE R, REG_DEPENDENCY D WHERE D.PARENT_RID=?" +
            //            " AND D.CHILD_RID=R.RID ORDER BY R.PATH";
            //    PreparedStatement ps = dataSource.getConnection().prepareStatement(sql);
            //    ps.setString(1, getId());
            //    List<String> childPathList = new ArrayList();
            //    ResultSet results = ps.executeQuery();
            //    int current = 0;
            //    while (results.next()) {
            //        String childID = results.getString("RID");
            //        if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
            //            if (current >= start && (pageLen == -1 || current < start + pageLen))
            //                childPathList.add(results.getString("PATH"));
            //            current++;
            //        }
            //    }
            //    if (current < start) {
            //        throw new RegistryException("Didn't have enough results to start at #" + start);
            //    }
            //    ps.close();
            //    return childPathList.toArray(new String[childPathList.size()]);
            //
            //} catch (Exception e) {
            //
            //    String msg = "Failed to get child paths of resource " +
            //            getPath() + ". " + e.getMessage();
            //    log.error(msg, e);
            //    throw new RegistryException(msg, e);
            //}

        } else {

            if (content instanceof String[]){

                List<String> childPathList = new ArrayList <String> ();
                String childNodes [] = (String[])content;
                int limit = start + pageLen;
                if (start > childNodes.length) {
                    return new String[0];
                }
                if (limit > childNodes.length) {
                    limit = childNodes.length;
                }
                for (int i = start ; i < limit; i++) {
                    String childNode = childNodes[i];
                    childPathList.add(childNode);
                }
                return childPathList.toArray(new String[childPathList.size()]);
            }
        }
        return new String[0];
    }

    public List<Resource> getChildResources() {
        return null;
    }

    public int getChildCount() throws RegistryException {

        if (childCount != -1) {
            return childCount;

        } else if (content != null && content instanceof String[]) {

            String[] childPaths = (String[]) content;
            return childPaths.length;

        } else {

            if (Transaction.isStarted()) {

                childCount = resourceDAO.getChildCount(getId(), Transaction.getConnection());

            } else {

                Connection conn = null;

                try {
                    conn = dataSource.getConnection();
                    childCount = resourceDAO.getChildCount(getId(), conn);

                } catch (SQLException e) {

                    String msg = "Failed to get the child count of resource " +
                            getPath() + ". " + e.getMessage();
                    log.error(msg, e);
                    throw new RegistryException(msg, e);

                } finally {

                    if (conn != null) {
                        try {
                            conn.close();
                        } catch (SQLException e) {
                            log.error("Failed to close the database connection used to " +
                                    "get child count of collection " + getPath());
                        }
                    }
                }
            }
        }

        return childCount;

        //if (content == null) {
        //    String sql = "SELECT COUNT(*) AS CHILD_COUNT FROM REG_DEPENDENCY WHERE PARENT_RID=?";
        //    try {
        //        PreparedStatement ps = dataSource.getConnection().prepareStatement(sql);
        //        ps.setString(1, getId());
        //        ResultSet results = ps.executeQuery();
        //        if (results.next()) {
        //            childCount = results.getInt("CHILD_COUNT");
        //            return childCount;
        //        }
        //    } catch (SQLException e) {
        //        e.printStackTrace();
        //    }
        //}  else {
        //    if ( content instanceof String[]){
        //        return ((String[])content).length;
        //    }
        //}
        //return 0;
    }


    public void setChildCount(int count) {
        childCount = count;
    }


    /**
     * Collection's content is a string array, which contains paths of its children. These paths
     * are loaded on demand to increase performance. It is recommended to use getChildren(...)
     * method to get child paths of a collection, which provides pagination. Calling this method
     * will load all child paths.
     *
     * @return String array of child paths.
     * @throws RegistryException On any error.
     */
    public Object getContent() throws RegistryException {

        if (content == null) {

            List <String> childPathList = null;

            if (Transaction.isStarted()) {
                childPathList = resourceDAO.getChildPaths(getId(), Transaction.getConnection());
            } else {

                Connection conn = null;

                try {
                    conn = dataSource.getConnection();
                    childPathList = resourceDAO.getChildPaths(getId(), conn);

                } catch (SQLException e) {

                    String msg = "Failed to get child paths of " +
                            getPath() + ". " + e.getMessage();
                    log.error(msg, e);
                    throw new RegistryException(msg, e);

                } finally {

                    if (conn != null) {
                        try {
                            conn.close();
                        } catch (SQLException e) {
                            log.error("Failed to close the database connection opened for " +
                                    "accessing the child paths of " + getPath(), e);
                        }
                    }
                }
            }

            content = childPathList.toArray(new String[childPathList.size()]);
        }

        return content;
    }
}
