/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Hits;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.wso2.registry.*;
import org.wso2.registry.Collection;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.exceptions.ResourceNotFoundException;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.jdbc.dao.*;
import org.wso2.registry.jdbc.handlers.RequestContext;
import org.wso2.registry.jdbc.queries.QueryProcessorManager;
import org.wso2.registry.session.CurrentSession;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.utils.AuthorizationUtils;
import org.wso2.registry.utils.RegistryUtils;
import org.wso2.registry.utils.VersionedPath;

import java.util.*;
import java.io.IOException;

/**
 * This class contains JDBC based implementations of all Registry API methods. This invokes handlers
 * through the HandlerManager for all the methods that allow handlers. Once the handler invocation
 * is complete and if the processing is not completed, it handles the request using the generic
 * procedure. For the methods that do not allow handlers, it simply processes the request using the
 * database layer.
 *
 * A valid database connection has to be available in the Transaction's thread local varaible,
 * when invoking methods of this class. Mehtods of this class neither start nor commit
 * the transactions when performing the operations. The caller of this class should manage database
 * transactions outside this class.
 *
 * There is only one instance of this class per registry instance. This instance can be invoked by
 * many threads. So it has to be thread safe.
 */
public class BasicRegistry implements Registry {

    public static final Log log = LogFactory.getLog(BasicRegistry.class);

    /**
     * Reference to the RegistryContext instance.
     */
    private RegistryContext registryContext;

    /**
     * Repository instance. This is used to handle basic resource storage operations of current
     * versions.
     */
    private Repository repository;

    /**
     * VersionRepository instance. This is used to handle resource storage operations of
     * archived (old) versions.
     */
    private VersionRepository versionRepository;

    /**
     * This is used to execute custom queries.
     */
    private QueryProcessorManager queryProcessorManager;

    /**
     * This is used to interact with the database layer to perform resource related operatons.
     */
    private ResourceDAO resourceDAO = new ResourceDAO();

    /**
     * This is used to interact with the database layer to perform associations related operations.
     */
    private AssociationDAO associationDAO = new AssociationDAO();

    /**
     * This is used to interact with the database layer to perform activity logging related
     * operations.
     */
    private LogsDAO logsDAO = new LogsDAO();

    /**
     * Instantiates the BasicRegistry instance. Objects passed to the constructor should be shared
     * between JDBCRegisty and BasicRegistry instances.
     *
     * @param registryContext RegistryContext instance containing the registry configuration.
     * @param repository Repository instance.
     * @param versionRepository VersionRepository instance.
     * @param queryProcessorManager QueryProcessorManager instance.
     */
    public BasicRegistry(RegistryContext registryContext,
                         Repository repository,
                         VersionRepository versionRepository,
                         QueryProcessorManager queryProcessorManager) {
        this.registryContext = registryContext;
        this.repository = repository;
        this.versionRepository = versionRepository;
        this.queryProcessorManager = queryProcessorManager;
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void beginTransaction() throws RegistryException {
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void commitTransaction() throws RegistryException {
    }

    /**
     * BasicRegistry does not implement transaction related methods as transactions has to be
     * handled outside of the BasicRegistry.
     */
    public void rollbackTransaction() throws RegistryException {
    }

    public Resource newResource() throws RegistryException {
        ResourceImpl resource = new ResourceImpl();
        resource.setAuthorUserName(CurrentSession.getUser());
        return resource;
    }

    public Collection newCollection() throws RegistryException {
        CollectionImpl coll = new CollectionImpl();
        coll.setAuthorUserName(CurrentSession.getUser());
        return coll;
    }

    public RegistryContext getRegistryContext() {
        return registryContext;
    }

    public void setRegistryContext(RegistryContext registryContext) {
        this.registryContext = registryContext;
    }

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws org.wso2.registry.exceptions.RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Resource get(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);

        // check if this path refers to a resource referred by a URL query (e.g. comment)

        RequestContext requestContext = new RequestContext(this, repository);
        requestContext.setResourcePath(resourcePath);

        Resource resource = registryContext.getHandlerManager().get(requestContext);

        // resource may have been fetched from the repository, to be used by handlers. if it is
        // done, it has to be stored in the request context. we can just use that resource
        // without fetching it again from the repository.
        if (resource == null) {
            resource = requestContext.getResource();
        }

        if (resource == null) {
            VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
            if (versionedPath.getVersion() == -1) {
                resource = repository.get(resourcePath.getPath());
            } else {
                resource = versionRepository.get(versionedPath);
            }
        }

        if (resource == null) {
            throw new ResourceNotFoundException(path);
        }

        return resource;
    }

    /**
     * Retrives a collection from the registry. This method supports the pagination of the
     * collection. Starting child entry number and number of entries per page can be specified.
     *
     * @param path Path of the collection
     * @param start Starting child entry number
     * @param pageSize Number of entries per page
     * @return Collection instance filled with specified subset of child paths
     * @throws RegistryException
     */
    public Collection get(String path, int start, int pageSize) throws RegistryException {

        Collection collection;

        ResourcePath resourcePath = new ResourcePath(path);
        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);
        if (versionedPath.getVersion() == -1) {
            collection = repository.get(resourcePath.getPath(), start, pageSize);
        } else {
            collection = versionRepository.get(versionedPath, start, pageSize);
        }

        return collection;
    }

    /**
     * Checks whether a resource exists in the registry.
     *
     * @param path Path of resource
     * @return true if resource exists, false otherwise.
     * @throws RegistryException
     */
    public boolean resourceExists(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);
        VersionedPath versionedPath = RegistryUtils.getVersionedPath(resourcePath);

        return versionedPath.getVersion() == -1 &&
                repository.resourceExists(resourcePath.getPath());
    }

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource.
     *
     * @param suggestedPath the path which we'd like to use for the new resource.
     * @param resource Resource instance for the new resource
     * @return the actual path that the server chose to use for our Resource
     * @throws RegistryException This exception is thrown for all database layer and handler
     * exceptions.
     */
    public synchronized String put(String suggestedPath, Resource resource)
            throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(suggestedPath);

        RequestContext requestContext = new RequestContext(this, repository);
        requestContext.setResourcePath(resourcePath);
        requestContext.setResource(resource);

        if (!RegistryConstants.ROOT_PATH.equals(resourcePath.getPath())) {
            registryContext.getHandlerManager().putChild(requestContext);
        }

        registryContext.getHandlerManager().put(requestContext);

        String actualPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            actualPath = suggestedPath;
            repository.put(resourcePath.getPath(), resource);

            // we are always creating versions for resources (files), if the resource has changed.
            if (!(resource instanceof Collection) &&
                    ((ResourceImpl) resource).isVersionableChange()) {
                createVersion(actualPath);
            }
        }

        if (actualPath == null) {
            return suggestedPath;
        } else {
            return actualPath;
        }
    }

    /**
     * Creates a new resource by importing the reosurce content from the given source URL.
     *
     * @param suggestedPath Path to put the resource.
     * @param sourceURL URL to import resource content.
     * @param metadata Metadata for the new resource.
     * @return Path where the resource is stored. Note that this may differ from the suggestedPath
     * as some modifications to the suggestedPath will be done if it does not comply with Registry
     * standards.
     * @throws RegistryException
     */
    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(suggestedPath);

        RequestContext importChildContext = new RequestContext(this, repository);
        ResourcePath parentPath =
                new ResourcePath(RegistryUtils.getParentPath(resourcePath.getPath()));
        importChildContext.setResourcePath(parentPath);
        registryContext.getHandlerManager().importChild(importChildContext);

        RequestContext requestContext = new RequestContext(this, repository);
        requestContext.setResourcePath(resourcePath);
        requestContext.setSourceURL(sourceURL);
        requestContext.setResource(metadata);

        registryContext.getHandlerManager().importResource(requestContext);
        String savedPath = requestContext.getActualPath();

        if (!requestContext.isProcessingComplete()) {
            savedPath = suggestedPath;

            // if some handlers have updated the metadata *without completing the request* we should
            // capture the updated metadata here.
            if (requestContext.getResource() != null) {
                metadata = requestContext.getResource();
            }

            repository.importResource(resourcePath.getPath(), sourceURL, metadata);
            createVersion(resourcePath.getPath());
        }

        logsDAO.addLog(savedPath, CurrentSession.getUser(), LogEntry.UPDATE, null);

        return suggestedPath;
    }

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.registry.exceptions.RegistryException
     *          : if the user is not authorized
     */
    public synchronized void delete(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);
        RequestContext requestContext = new RequestContext(this, repository);
        requestContext.setResourcePath(resourcePath);

        registryContext.getHandlerManager().delete(requestContext);

        if (!requestContext.isProcessingComplete()) {
            repository.delete(resourcePath.getPath());
        }

        logsDAO.addLog(
                resourcePath.getPath(), CurrentSession.getUser(), LogEntry.DELETE_RESOURCE, null);
    }

    public String rename(String currentPath, String newName) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(currentPath);
        String newPath  = repository.rename(resourcePath, newName);
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.RENAME, newName);

        return newPath;
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        ResourcePath currentResourcePath = new ResourcePath(currentPath);
        String movedPath  = repository.move(currentResourcePath, newPath);
        logsDAO.addLog(currentPath, CurrentSession.getUser(), LogEntry.MOVE, newPath);

        return movedPath;
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        String copiedPath  = repository.copy(sourceResourcePath, targetResourcePath);
        logsDAO.addLog(sourcePath, CurrentSession.getUser(), LogEntry.COPY, targetPath);

        return copiedPath;
    }

    public void createVersion(String path) throws RegistryException {

        ResourcePath resourcePath = new ResourcePath(path);
        versionRepository.createSnapshot(resourcePath);
    }

    public String[] getVersions(String path) throws RegistryException {

        return versionRepository.getVersions(path);
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        ResourcePath versionedResourcePath = new ResourcePath(versionPath);
        versionRepository.restoreVersion(versionedResourcePath);
    }

    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        if (repository.resourceExists(sourceResourcePath.getPath())) {
            sourcePath = sourceResourcePath.getPath();
        }

        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        if (repository.resourceExists(targetResourcePath.getPath())) {
            targetPath = targetResourcePath.getPath();
        }

        associationDAO.addAssociation(sourcePath, targetPath, associationType);
    }

    public void removeAssociation(
            String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath sourceResourcePath = new ResourcePath(sourcePath);
        if (repository.resourceExists(sourceResourcePath.getPath())) {
            sourcePath = sourceResourcePath.getPath();
        }

        ResourcePath targetResourcePath = new ResourcePath(targetPath);
        if (repository.resourceExists(targetResourcePath.getPath())) {
            targetPath = targetResourcePath.getPath();
        }

        associationDAO.removeAssociation(sourcePath, targetPath, associationType);
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (repository.resourceExists(processedPath.getPath())) {
            resourcePath = processedPath.getPath();
        }

        return associationDAO.getAllAssociations(resourcePath);
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        // Source and target of associations may or may not be resources in the registry. If they
        // don't refer to a resource, they can contain any string value. But if they refer to
        // resources, values should be proper resource paths.

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (repository.resourceExists(processedPath.getPath())) {
            resourcePath = processedPath.getPath();
        }

        return associationDAO.getAllAssociationsForType(resourcePath, associationType);
    }


    ////////////////////////////////////////////////////////
    // Tagging
    ////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined in
     * the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag Tag. Any string can be used for the tag.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void applyTag(String resourcePath, String tag) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to apply tag to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        // break the comma separated words into multiple tags
        String[] tags = tag.split(",");

        String resourceID;
        if ((resourceID = ResourceDAO.getResourceID(resourcePath)) == null) {
            String msg = "Failed to apply tag " + tag + " on resource " + resourcePath +
                    ". Resource " + resourcePath + " does not exist.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String userName = CurrentSession.getUser();

        if (!AuthorizationUtils.authorize(resourceID, ActionConstants.GET)) {
            String msg = "Failed to apply tag " + tag + " on resource " + resourcePath +
                    ". User " + userName + " is not authorized to read the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        for (int i = 0; i < tags.length; i++) {

            tags[i] = tags[i].trim();

            if (tags[i].length() == 0 || tags[i].equals(" ")) {
                continue;
            }

            if (TagsDAO.taggingExists(tags[i], resourceID, userName)) {
                // Already there, don't bother doing it again.
                return;
            }

            TagsDAO.addTagging(tags[i], resourceID, userName);
            logsDAO.addLog(resourcePath, userName, LogEntry.TAG, tags[i]);
        }
    }

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     */
    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        // break the tags from spaces
        String[] tags = tag.trim().split(",");

        for (int i = 0; i < tags.length; i++) {
            tags[i] = tags[i].trim();
        }

        List taggedPaths = new ArrayList();

        List pathList = TagsDAO.getPathsWithAnyTag(tags);
        Iterator iPaths = pathList.iterator();
        while (iPaths.hasNext()) {
            String path = (String)iPaths.next();

            TaggedResourcePath taggedResourcePath = new TaggedResourcePath();
            taggedResourcePath.setResourcePath(path);

            for (String currentTag : tags) {
                long count = TagsDAO.getTagCount(path, currentTag);
                taggedResourcePath.addTagCount(currentTag, count);
            }

            taggedPaths.add(taggedResourcePath);
        }

        return (TaggedResourcePath[])taggedPaths
                .toArray(new TaggedResourcePath[taggedPaths.size()]);
    }

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param resourcePath Path of the resource
     * @return Tags
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public Tag[] getTags(String resourcePath) throws RegistryException {

        resourcePath = RegistryUtils.prepareGeneralPath(resourcePath);
        resourcePath = getCurrentPath(resourcePath);

        return TagsDAO.getTagsWithCount(resourcePath);
    }

    public void removeTag(String path, String tag) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(path);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to remove tag from the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        path = processedPath.getPath();

        path = RegistryUtils.prepareGeneralPath(path);

        String user = CurrentSession.getUser();

        boolean adminUser = false;

        // check is the user belongs to the admin role
        try {
            UserRealm userRealm = CurrentSession.getRealm();
            String[] roles = userRealm.getUserStoreReader().getUserRoles(user);
            if (RegistryUtils.containsString(RegistryConstants.ADMIN_ROLE, roles)) {
                adminUser = true;
            }

        } catch (UserStoreException e) {

            String msg = "Failed to get roles of the current user. " +
                    "User will be considered as non-admin user.\n" + e.getMessage();
            log.info(msg, e);
            adminUser = false;
        }

        // check if the user is the admin user
        // TODO - do we really need to do this check?  Won't this user always be in the admin role?
        if (RegistryConstants.ADMIN_USER.equals(user)) {
            adminUser = true;
        }

        if (adminUser) {
            TagsDAO.removeTag(path, tag);
            return;
        }

        String author = resourceDAO.getResourceAuthor(path);
        if (user.equals(author)) {
            TagsDAO.removeTag(path, tag);
        } else {
            TagsDAO.removeTag(path, tag, user);
        }
    }

    ////////////////////////////////////////////////////////
    // Comments
    ////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment      Comment instance for the new comment.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized String addComment(String resourcePath, Comment comment)
            throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to add comment to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        String userName = CurrentSession.getUser();

        String resourceID;
        if ((resourceID = ResourceDAO.getResourceID(resourcePath)) == null) {
            String msg = "Failed to add comment " + comment.getText() + " to resource " +
                    resourcePath + ". Resource " + resourcePath + " does not exist.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        if (!AuthorizationUtils.authorize(resourceID, ActionConstants.GET)) {
            String msg = "Failed to add comment " + comment.getText() + " on resource " +
                    resourcePath + ". User " + userName +
                    " is not authorized to read the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        String commentPath = CommentsDAO.addComment(resourceID, userName, comment);
        commentPath = resourcePath + RegistryConstants.URL_SEPARATOR + "comments:" + commentPath;
        logsDAO.addLog(
                resourcePath, userName, LogEntry.COMMENT, comment.getText());

        return commentPath;
    }


    public void editComment(String commentPath, String text) throws RegistryException {

        String[] parts = commentPath.split(RegistryConstants.URL_SEPARATOR);
        String commentPart = parts[1];

        if (parts.length == 2 && commentPart.startsWith("comments:")) {
            String commentId = parts[1].substring(9);
            CommentsDAO.updateComment(Long.parseLong(commentId), text);
        }
    }

    /**
     * Returns all comments for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Array of Comment objects.
     */
    public Comment[] getComments(String resourcePath) throws RegistryException {

        resourcePath = getCurrentPath(resourcePath);

        return CommentsDAO.getComments(resourcePath);
    }

    ////////////////////////////////////////////////////////
    // Ratings
    ////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating       rating value.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void rateResource(String resourcePath, int rating)
            throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to apply rating to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        String userName = CurrentSession.getUser();

        String resourceID = ResourceDAO.getResourceID(resourcePath);

        if (resourceID != null) {

            if (!AuthorizationUtils.authorize(resourceID, ActionConstants.GET)) {
                String msg = "Failed to rate resource " + resourcePath + " with rating " + rating +
                        ". User " + userName + " is not authorized to read the resource.";
                log.error(msg);
                throw new RegistryException(msg);
            }

            if (RatingsDAO.ratingExists(resourceID, userName)) {

                RatingsDAO.updateRating(resourceID, userName, rating);

                log.info("Updated the rating on the resource " +
                        resourcePath + " by user " + userName);
            } else {
                RatingsDAO.addRating(resourceID, userName, rating);
            }

            logsDAO.addLog(resourcePath,
                    userName, LogEntry.RATING, Integer.toString(rating));

        } else {
            String msg = Messages.getMessage("rate.on.null.artfact", resourcePath);
            log.error(msg);
            throw new RegistryException(msg);
        }
    }

    /**
     * Returns the average rating (from all users) for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return average rating.
     */
    public float getAverageRating(String resourcePath) throws RegistryException {

        return RatingsDAO.getAverageRating(resourcePath);
    }

    public int getRating(String path, String userName) throws RegistryException {

        return RatingsDAO.getRating(path, userName);
    }

    //////////////////////////////////////////////////////////////////////////////
    // Extensible searching API
    //////////////////////////////////////////////////////////////////////////////

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        Resource query = get(path);

        return queryProcessorManager.executeQuery(this, query, parameters);
    }


    private String getCurrentPath(String resourcePath) {
        String currentPath = resourcePath;
        if (resourcePath.indexOf("?") > 0) {
            currentPath = resourcePath.split("\\?")[0];
        } else if (resourcePath.indexOf(RegistryConstants.URL_SEPARATOR) > 0) {
            currentPath = resourcePath.split("\\;")[0];
        }

        return currentPath;
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from, Date to, boolean recentFirst) throws RegistryException {

        List logEntryList  =
                logsDAO.getLogs(resourcePath, action, userName, from, to, recentFirst);

        LogEntry[] logEntries = new LogEntry[logEntryList.size()];
        for (int i = 0; i < logEntryList.size(); i++) {
            logEntries[i] = (LogEntry)logEntryList.get(i);
        }

        return logEntries;
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        LogEntryCollection logEntryCollection = new LogEntryCollection();
        logEntryCollection.setLogCount(logsDAO.getLogsCount(resourcePath,action,userName,from,to,recentFirst));
        return logEntryCollection;
    }

    public String[] getAvailableAspects() {
        return registryContext.getAspectNames();
    }

    public void associateAspect(String resourcePath, String name) throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to associate aspect to the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        resourcePath = processedPath.getPath();

        //TODO need to do the security validation here
        Resource resource = get(resourcePath);
        Aspect aspect = getAspect(name);
        if (aspect == null) {
            throw new RegistryException("Couldn't find aspect '" + name + "'");
        }
        aspect.associate(resource, this);
        resource.addAspect(name);
        put(resource.getPath(), resource);
    }

    private Aspect getAspect(String name) throws RegistryException {
        return registryContext.getAspect(name);
    }

    //public void addAspect(String name, Aspect aspect) {
    //    aspects.put(name, aspect);
    //}

    public Aspect getResourceAspect(Resource resource, String aspectName)
            throws RegistryException {

        Aspect aspect = getAspect(aspectName);
        if (aspect == null) {
            throw new RegistryException("Aspect '" + aspectName + "' is not registered!");
        }

        // Confirm this aspect is associated with this Resource
        List<String> resourceAspects = resource.getAspects();
        if (resourceAspects == null || !resourceAspects.contains(aspectName)) {
            throw new RegistryException("Resource at '" + resource.getPath() +
                    "' not associated with aspect '" + aspectName + "'");
        }

        return aspect;
    }

    public void invokeAspect(String resourcePath, String aspectName, String action)
            throws RegistryException {

        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to invoke aspect of the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourcePath = processedPath.getPath();

        Resource resource = repository.get(resourcePath);

        Aspect aspect = getResourceAspect(resource, aspectName);
        RequestContext context = new RequestContext(this, repository);
        context.setResource(resource);

//        List aspectNames = resource.getPropertyValues(Aspect.AVAILABLE_ASPECTS);
//        if (aspectNames == null) {
//            throw new RegistryException("No aspect are associated with the resource");
//        }

        aspect.invoke(context, action);
        put(resourcePath, resource);
    }

    public String[] getAspectActions(String resourcePath, String aspectName)
            throws RegistryException {
        
        ResourcePath processedPath = new ResourcePath(resourcePath);
        if (!processedPath.isCurrentVersion()) {
            String msg = "Failed to get aspect actions of the resource " + processedPath +
                    ". Given path refers to an archived version of the resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }
        resourcePath = processedPath.getPath();

        Resource resource = repository.get(resourcePath);
        Aspect aspect = getResourceAspect(resource, aspectName);

        RequestContext context = new RequestContext(this, repository);
        context.setResource(resource);
        return aspect.getAvailableActions(context);
    }

    public Collection searchContent (String keywords) throws RegistryException{
        try{
            Searcher searcher = new IndexSearcher(RegistryContext.getSingleton().getRamDir());
            Query query = new QueryParser("content", new StandardAnalyzer()).parse(keywords);
            Hits hits = searcher.search(query);
            Collection collection = new CollectionImpl();
            String [] paths = new String [hits.length()];
            for (int i = 0; i < hits.length(); i++) {
                String id = hits.doc(i).get("id");
                String path = resourceDAO.getResourcePath(id);
                paths[i] = path;
            }
            collection.setContent(paths);
            return collection;
        } catch (IOException e) {
            String msg = "Failed to search content";
            log.error(msg);
            throw new RegistryException(msg);
        } catch (ParseException e) {
            String msg = "Failed to parse the query";
            log.error(msg);
            throw new RegistryException(msg);
        }
    }
}
