/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.LogEntry;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.DatabaseConstants;
import org.wso2.registry.jdbc.utils.Transaction;

import javax.sql.DataSource;
import java.sql.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class LogsDAO {

    private static final Log log = LogFactory.getLog(LogsDAO.class);

    public void addLog(String resourcePath,
                       String userName,
                       int action,
                       String actionData)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_LOG (PATH, USER_ID, LOGGED_TIME, ACTION, ACTION_DATA) " +
                         "VALUES (?, ?, ?, ?, ?)";

            PreparedStatement s = conn.prepareStatement(sql);
            s.setString(1, resourcePath);
            s.setString(2, userName);
            s.setTimestamp(3, new Timestamp(System.currentTimeMillis()));
            s.setInt(4, action);
            s.setString(5, actionData);

            s.executeUpdate();
            s.close();

        } catch (SQLException e) {

            String msg = "Failed to add log entry for resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public List getLogs(String resourcePath, int action,
                        String userName, Date from, Date to, boolean descending)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT PATH, USER_ID, LOGGED_TIME, ACTION, ACTION_DATA FROM " +
                     "REG_LOG";

        boolean queryStarted = false;

        sql = addWherePart(resourcePath, queryStarted, sql, userName, from, to, action);

        if (descending) {
            sql = sql + " ORDER BY LOGGED_TIME DESC";
        }

        try {
            PreparedStatement s = conn.prepareStatement(sql);

            int paramNumber = 1;

            if (resourcePath != null) {
                s.setString(paramNumber, resourcePath);
                paramNumber++;
            }

            if (userName != null) {
                s.setString(paramNumber, userName);
                paramNumber++;
            }

            if (from != null) {
                s.setTimestamp(paramNumber, new Timestamp(from.getTime()));
                paramNumber++;
            }

            if (to != null) {
                s.setTimestamp(paramNumber, new Timestamp(to.getTime()));
                paramNumber++;
            }

            if (action != -1) {
                s.setInt(paramNumber, action);
            }

            ResultSet results = s.executeQuery();

            List resultList = new ArrayList();
            while (results.next()) {
                LogEntry logEntry = new LogEntry();
                logEntry.setResourcePath(results.getString(DatabaseConstants.PATH_FIELD));
                logEntry.setUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
                logEntry.setDate(
                        new Date(results.getTimestamp(DatabaseConstants.LOGGED_TIME_FIELD).getTime()));
                logEntry.setAction(results.getInt(DatabaseConstants.ACTION_FIELD));
                logEntry.setActionData(results.getString(DatabaseConstants.ACTION_DATA_FIELD));

                resultList.add(logEntry);
            }
            s.close();

            return resultList;

        } catch (SQLException e) {

            String msg = "Failed to get logs. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private String addWherePart(String resourcePath,
                                boolean queryStarted,
                                String sql,
                                String userName,
                                Date from,
                                Date to,
                                int action) {
        if (resourcePath != null) {
            if (queryStarted) {
                sql = sql + " AND PATH=?";
            } else {
                sql = sql + " WHERE PATH=?";
                queryStarted = true;
            }
        }

        if (userName != null) {
            if (queryStarted) {
                sql = sql + "  AND USER_ID=?";
            } else {
                sql = sql + "  WHERE USER_ID=?";
                queryStarted = true;
            }
        }

        if (from != null) {
            if (queryStarted) {
                sql = sql + " AND LOGGED_TIME>?";
            } else {
                sql = sql + " WHERE LOGGED_TIME>?";
                queryStarted = true;
            }
        }

        if (to != null) {
            if (queryStarted) {
                sql = sql + " AND LOGGED_TIME<?";
            } else {
                sql = sql + " WHERE LOGGED_TIME<?";
                queryStarted = true;
            }
        }

        if (action != -1) {
            if (queryStarted) {
                sql = sql + " AND ACTION=?";
            } else {
                sql = sql + " WHERE ACTION=?";
                queryStarted = true;
            }
        }
        return sql;
    }

    public LogEntry[] getLogs(String resourcePath,
                        int action,
                        String userName,
                        Date from,
                        Date to,
                        boolean descending ,
                        int start,
                        int pageLen ,
                        DataSource dataSource)
            throws RegistryException {
        String sql = "SELECT PATH, USER_ID, LOGGED_TIME, ACTION, ACTION_DATA FROM " +
                     "REG_LOG";
        boolean queryStarted = false;

        sql = addWherePart(resourcePath, queryStarted, sql, userName, from, to, action);

        if (descending) {
            sql = sql + " ORDER BY LOGGED_TIME DESC";
        }

        try {
            Connection conn = dataSource.getConnection();
            PreparedStatement s = conn.prepareStatement(sql);

            int paramNumber = 1;

            if (resourcePath != null) {
                s.setString(paramNumber, resourcePath);
                paramNumber++;
            }

            if (userName != null) {
                s.setString(paramNumber, userName);
                paramNumber++;
            }

            if (from != null) {
                s.setTimestamp(paramNumber, new Timestamp(from.getTime()));
                paramNumber++;
            }

            if (to != null) {
                s.setTimestamp(paramNumber, new Timestamp(to.getTime()));
                paramNumber++;
            }

            if (action != -1) {
                s.setInt(paramNumber, action);
            }

            ResultSet results = s.executeQuery();

            List<LogEntry> resultList = new ArrayList();
            int current = 0;
            while (results.next()) {
                if (current >= start && (pageLen == -1 || current < start + pageLen)) {
                    LogEntry logEntry = new LogEntry();
                    logEntry.setResourcePath(results.getString(DatabaseConstants.PATH_FIELD));
                    logEntry.setUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
                    logEntry.setDate(
                            new Date(results.getTimestamp(DatabaseConstants.LOGGED_TIME_FIELD).getTime()));
                    logEntry.setAction(results.getInt(DatabaseConstants.ACTION_FIELD));
                    logEntry.setActionData(results.getString(DatabaseConstants.ACTION_DATA_FIELD));

                    resultList.add(logEntry);
                }
                current++;
            }
            s.close();
            return resultList.toArray(new LogEntry[resultList.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get logs. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

     public LogEntry[] getLogs(String resourcePath,
                        int action,
                        String userName,
                        Date from,
                        Date to,
                        boolean descending ,
                        DataSource dataSource)
            throws RegistryException {

        String sql = "SELECT PATH, USER_ID, LOGGED_TIME, ACTION, ACTION_DATA FROM " +
                     "REG_LOG";

        boolean queryStarted = false;

        sql = addWherePart(resourcePath, queryStarted, sql, userName, from, to, action);

        if (descending) {
            sql = sql + " ORDER BY LOGGED_TIME DESC";
        }

        try {
            Connection conn = dataSource.getConnection();
            PreparedStatement s = conn.prepareStatement(sql);

            int paramNumber = 1;

            if (resourcePath != null) {
                s.setString(paramNumber, resourcePath);
                paramNumber++;
            }

            if (userName != null) {
                s.setString(paramNumber, userName);
                paramNumber++;
            }

            if (from != null) {
                s.setTimestamp(paramNumber, new Timestamp(from.getTime()));
                paramNumber++;
            }

            if (to != null) {
                s.setTimestamp(paramNumber, new Timestamp(to.getTime()));
                paramNumber++;
            }

            if (action != -1) {
                s.setInt(paramNumber, action);
            }

            ResultSet results = s.executeQuery();

            List<LogEntry> resultList = new ArrayList();
            while (results.next()) {
                    LogEntry logEntry = new LogEntry();
                    logEntry.setResourcePath(results.getString(DatabaseConstants.PATH_FIELD));
                    logEntry.setUserName(results.getString(DatabaseConstants.USER_ID_FIELD));
                    logEntry.setDate(
                            new Date(results.getTimestamp(DatabaseConstants.LOGGED_TIME_FIELD).getTime()));
                    logEntry.setAction(results.getInt(DatabaseConstants.ACTION_FIELD));
                    logEntry.setActionData(results.getString(DatabaseConstants.ACTION_DATA_FIELD));

                    resultList.add(logEntry);
            }
            s.close();
            return resultList.toArray(new LogEntry[resultList.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get logs. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public int getLogsCount(String resourcePath,
                            int action,
                            String userName,
                            Date from,
                            Date to,
                            boolean descending)
            throws RegistryException {
        int count =0 ;

        Connection conn = Transaction.getConnection();

        String sql = "SELECT COUNT(*) AS LOG_COUNT FROM REG_LOG";

        boolean queryStarted = false;

        sql = addWherePart(resourcePath, queryStarted, sql, userName, from, to, action);

        try {
            PreparedStatement s = conn.prepareStatement(sql);

            int paramNumber = 1;

            if (resourcePath != null) {
                s.setString(paramNumber, resourcePath);
                paramNumber++;
            }

            if (userName != null) {
                s.setString(paramNumber, userName);
                paramNumber++;
            }

            if (from != null) {
                s.setTimestamp(paramNumber, new Timestamp(from.getTime()));
                paramNumber++;
            }

            if (to != null) {
                s.setTimestamp(paramNumber, new Timestamp(to.getTime()));
                paramNumber++;
            }

            if (action != -1) {
                s.setInt(paramNumber, action);
            }

            ResultSet results = s.executeQuery();

            if (results.next()) {
                count =  results.getInt("LOG_COUNT");
            }
            s.close();


        } catch (SQLException e) {

            String msg = "Failed to get logs. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        return count;
    }
}
