/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.*;
import org.wso2.registry.Collection;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.dataobjects.ResourceDO;
import org.wso2.registry.jdbc.filecache.FileManager;
import org.wso2.registry.jdbc.utils.Transaction;
import org.wso2.registry.jdbc.utils.ResourceVersionHolder;
import org.wso2.registry.utils.AuthorizationUtils;

import java.io.*;
import java.sql.*;
import java.util.*;

public class ResourceVersionDAO {

    private static Log log = LogFactory.getLog(ResourceVersionDAO.class);

    public long addSnapshot(String snapshotRootID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_SNAPSHOT (ROOT_ID) VALUES (?)";
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, snapshotRootID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add snapshot for the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        String sql2 = "SELECT MAX(SNAPSHOT_ID) FROM REG_SNAPSHOT WHERE ROOT_ID=? GROUP BY ROOT_ID";
        try {
            PreparedStatement ps2 = conn.prepareStatement(sql2);
            ps2.setString(1, snapshotRootID);

            ResultSet result = ps2.executeQuery();
            long snapshotID = 0;
            if (result.next()) {
                snapshotID = result.getLong(1);
            }

            return snapshotID;

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot ID of the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public int getLatestResourceVersionNumber(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(VERSOIN) " +
                "FROM REG_RESOURCE_VERSION WHERE RID=? GROUP BY RID";

        int snapshotNumber = 0;
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                snapshotNumber = result.getInt(1);
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return snapshotNumber;
    }

    public int addResourceVersion(ResourceImpl resourceImpl, String contentVersionID)
            throws RegistryException {

        int versionNumber = getMaxResourceVersion(resourceImpl.getId()) + 1;

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_RESOURCE_VERSION (RID, VERSION, PATH, MEDIA_TYPE, COLLECTION, " +
                "CREATOR, CREATED_TIME, LAST_UPDATOR, LAST_UPDATED_TIME, DESCRIPTION, CONTENT_ID, " +
                "ASSOCIATED_SNAPSHOT_ID) " +
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            ps.setString(1, resourceImpl.getId());
            ps.setInt(2, versionNumber);
            ps.setString(3, resourceImpl.getPath());
            ps.setString(4, resourceImpl.getMediaType());
            ps.setInt(5, (resourceImpl instanceof Collection) ? 1: 0);
            ps.setString(6, resourceImpl.getAuthorUserName());
            ps.setTimestamp(7, new Timestamp(resourceImpl.getCreatedTime().getTime()));
            ps.setString(8, resourceImpl.getLastUpdaterUserName());
            ps.setTimestamp(9, new Timestamp(resourceImpl.getLastModified().getTime()));
            ps.setString(10, resourceImpl.getDescription());
            ps.setString(11, contentVersionID);
            ps.setLong(12, resourceImpl.getMatchingSnapshotID());

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add new version for resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return versionNumber;
    }

    public int getMaxResourceVersion(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(VERSION) FROM REG_RESOURCE_VERSION WHERE RID=? GROUP BY RID";

        int maxVersion = 0;
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                maxVersion = result.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to latest version number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return maxVersion;
    }

    public void addDependencies(String resourceID, int versionNumber, List<String> childIDs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_DEPENDENCY_VERSION (PARENT_RID, PARENT_VERSION, CHILD_RID) " +
                "VALUES (?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            Iterator <String> iChildIDs = childIDs.iterator();
            while(iChildIDs.hasNext()) {
                String childID = iChildIDs.next();

                ps.setString(1, resourceID);
                ps.setInt(2, versionNumber);
                ps.setString(3, childID);

                ps.executeUpdate();
                ps.clearParameters();
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add dependencies for version " + versionNumber +
                    " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void addProperties(ResourceImpl resource, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO REG_PROPERTY_VERSION (RID, VERSION, NAME, PROPERTY_VALUE) " +
                    "VALUES (?, ?, ?, ?)";

            PreparedStatement ps = conn.prepareStatement(sql);

            String resourceID = resource.getId();
            Properties props = resource.getProperties();
            if (props != null) {

                Iterator i = props.keySet().iterator();
                while (i.hasNext()) {
                    String name = (String)i.next();

                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        Iterator<String> iValues = propValues.iterator();
                        while (iValues.hasNext()) {
                            String value = iValues.next();

                            ps.setString(1, resourceID);
                            ps.setLong(2, version);
                            ps.setString(3, name);
                            ps.setString(4, value);

                            ps.executeUpdate();
                            ps.clearParameters();
                        }
                    }
                }
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to version properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void addSnapshotResource(long snapshotID, String resourceID, int versionNumber)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_SNAPSHOT_RESOURCE_VERSION (SNAPSHOT_ID, RID, VERSION) " +
                "VALUES (?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);
            ps.setInt(3, versionNumber);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add snapshot resource version for version " +
                    versionNumber + " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public Long[] getSnapshotIDs(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM REG_SNAPSHOT " +
                "WHERE ROOT_ID=(SELECT RID FROM REG_RESOURCE WHERE PATH=?) ORDER BY SNAPSHOT_ID DESC";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourcePath);

            ResultSet results = ps.executeQuery();
            List <Long> snapshotIDs = new ArrayList();
            while(results.next()) {
                long snapshotNumber = results.getLong("SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }
            ps.close();

            return snapshotIDs.toArray(new Long[snapshotIDs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public long getVersion(String resourceID, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE SNAPSHOT_ID=? AND RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);

            ResultSet result = ps.executeQuery();
            long version = -1;
            if (result.next()) {
                version = result.getLong("VERSION");
            }
            ps.close();

            return version;

        } catch (SQLException e) {

            String msg = "Failed to get version of resource " + resourceID +
                    " of snapshot " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public ResourceImpl get(String resourceID, long snapshotID, long version)
            throws RegistryException {

        ResourceImpl resourceImpl = getResource(resourceID, version);

        //if (resourceImpl instanceof Collection) {
        //    fillChildren(resourceImpl, snapshotID);
        //}

        fillResourceProperties(resourceImpl);

        return resourceImpl;
    }

    public CollectionImpl get(String resourceID, long snapshotID,
                              long version, int start, int pageLen)
            throws RegistryException {

        ResourceImpl resourceImpl = getResource(resourceID, version);

        if (!(resourceImpl instanceof CollectionImpl)) {
            String msg = "Child resource range can only be specified for collections. " +
                    resourceImpl.getPath() + " is not a collection.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        CollectionImpl collectionImpl = (CollectionImpl) resourceImpl;

        fillChildren(collectionImpl, snapshotID, start, pageLen);

        fillResourceProperties(collectionImpl);

        return collectionImpl;
    }

    private void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String propSQL = "SELECT NAME, PROPERTY_VALUE FROM REG_PROPERTY_VERSION WHERE RID=? AND VERSION=?";
            PreparedStatement ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());
            ResultSet results = ps.executeQuery();

            while (results.next()) {
                String name = results.getString("NAME");
                String value = results.getString("PROPERTY_VALUE");
                resourceImpl.addProperty(name, value);
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        resourceImpl.setPropertiesModified(false);
    }

    public int getChildCount(String resourceID, long versionNumber, Connection conn)
            throws RegistryException {

        try {

            int childCount = 0;

            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            ResultSet results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    childCount++;
                }
            }
            ps.close();

            return childCount;

        } catch (SQLException e) {

            String msg = "Failed to get child count of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void fillChildren(ResourceImpl resourceImpl, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List <String> childPathList = new ArrayList();
            ResultSet results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("CHILD_RID");

                String sql2 = "SELECT R.PATH FROM REG_RESOURCE_VERSION R " +
                        "WHERE R.RID=? AND R.VERSION=(" +
                        "SELECT S.VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                        "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

                PreparedStatement ps2 = conn.prepareStatement(sql2);
                ps2.setString(1, childID);
                ps2.setLong(2, snapshotID);
                ps2.setString(3, childID);

                ResultSet result2 = ps2.executeQuery();
                String path = "";
                if (result2.next()) {
                    path = result2.getString("PATH");
                }
                ps2.close();

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    String childPath = path + ";version:" + snapshotID;
                    childPathList.add(childPath);
                }
            }
            ps.close();

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            resourceImpl.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public List <String> getChildPaths(
            String resourceID, long snapshotID, long versionNumber, Connection conn)
            throws RegistryException {

        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            List <String> childPathList = new ArrayList <String> ();
            ResultSet results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("CHILD_RID");
                
                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    String sql2 = "SELECT R.PATH FROM REG_RESOURCE_VERSION R " +
                            "WHERE R.RID=? AND R.VERSION=(" +
                            "SELECT S.VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                            "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

                    PreparedStatement ps2 = conn.prepareStatement(sql2);
                    ps2.setString(1, childID);
                    ps2.setLong(2, snapshotID);
                    ps2.setString(3, childID);

                    ResultSet result2 = ps2.executeQuery();
                    String path = "";
                    if (result2.next()) {
                        path = result2.getString("PATH");
                    }
                    ps2.close();

                    String childPath = path + ";version:" + snapshotID;
                    childPathList.add(childPath);
                }
            }
            ps.close();

            return childPathList;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of snapshot " + snapshotID + " of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void fillChildren(ResourceImpl resourceImpl, long snapshotID, int start, int pageLen)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List <String> childPathList = new ArrayList();
            ResultSet results = ps.executeQuery();

            int resultCount = 0;
            int end = start + pageLen;
            while (results.next() && resultCount <= end) {

                String childID = results.getString("CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    if (resultCount >= start && resultCount < end) {

                        String sql2 = "SELECT R.PATH FROM REG_RESOURCE_VERSION R " +
                                "WHERE R.RID=? AND R.VERSION=(" +
                                "SELECT S.VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                                "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

                        PreparedStatement ps2 = conn.prepareStatement(sql2);
                        ps2.setString(1, childID);
                        ps2.setLong(2, snapshotID);
                        ps2.setString(3, childID);

                        ResultSet result2 = ps2.executeQuery();
                        String path = "";
                        if (result2.next()) {
                            path = result2.getString("PATH");
                        }
                        ps2.close();

                        String childPath = path + ";version:" + snapshotID;
                        childPathList.add(childPath);
                    }

                    resultCount++;
                }
            }
            ps.close();

            if (resultCount < start) {
                String msg = "Not enough child resources found to start at the child number " +
                        start + " of version " + snapshotID + " of resource " +
                        resourceImpl.getPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            resourceImpl.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public List <String> getChildPaths(
            ResourceImpl resourceImpl, long snapshotID, int start, int pageLen, Connection conn)
            throws RegistryException {

        try {
            //String sql = "SELECT R.RID, R.PATH FROM REG_RESOURCE_VERSION R, REG_DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List <String> childPathList = new ArrayList();
            ResultSet results = ps.executeQuery();

            int resultCount = 0;
            int end = start + pageLen;
            while (results.next() && resultCount <= end) {

                String childID = results.getString("CHILD_RID");

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {

                    if (resultCount >= start && resultCount < end) {

                        String sql2 = "SELECT R.PATH FROM REG_RESOURCE_VERSION R " +
                                "WHERE R.RID=? AND R.VERSION=(" +
                                "SELECT S.VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                                "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

                        PreparedStatement ps2 = conn.prepareStatement(sql2);
                        ps2.setString(1, childID);
                        ps2.setLong(2, snapshotID);
                        ps2.setString(3, childID);

                        ResultSet result2 = ps2.executeQuery();
                        String path = "";
                        if (result2.next()) {
                            path = result2.getString("PATH");
                        }
                        ps2.close();

                        String childPath = path + ";version:" + snapshotID;
                        childPathList.add(childPath);
                    }

                    resultCount++;
                }
            }
            ps.close();

            if (resultCount < start) {
                String msg = "Not enough child resources found to start at the child number " +
                        start + " of version " + snapshotID + " of resource " +
                        resourceImpl.getPath();
                log.error(msg);
                throw new RegistryException(msg);
            }

            return childPathList;

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private ResourceImpl getResource(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT RID, PATH, MEDIA_TYPE, COLLECTION, CREATOR, CREATED_TIME, " +
                    "LAST_UPDATOR, LAST_UPDATED_TIME, DESCRIPTION, CONTENT_ID " +
                    "FROM REG_RESOURCE_VERSION WHERE RID=? AND VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet result = ps.executeQuery();

            ResourceImpl resourceImpl = null;

            if (result.next()) {

                if (result.getInt("COLLECTION") == 0) {
                    resourceImpl = new ResourceVersionImpl();
                } else {
                    resourceImpl = new CollectionVersionImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setVersionNumber(version);

                resourceImpl.setId(result.getString("RID"));
                resourceImpl.setPath(result.getString("PATH"));
                resourceImpl.setMediaType(result.getString("MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("LAST_UPDATED_TIME"));
                resourceImpl.setDescription(result.getString("DESCRIPTION"));
                resourceImpl.setDbBasedContentID(result.getString("CONTENT_ID"));
            }

            ps.close();

            return resourceImpl;

        } catch (SQLException e) {

            String msg = "Failed to get resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public String addContentVersion(InputStream contentStream) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String contentID = null;

        if (contentStream != null) {

            contentID = UUID.randomUUID().toString();

            InputStream tempFileStream = null;
            try {

                //InputStream tempFileStream = new BufferedInputStream(new FileInputStream(tempFile));
                tempFileStream = FileManager.getInstance().
                        createFileBasedInputStream(contentID, contentStream);
                long contentLength = FileManager.getInstance().getContentLength(contentID);

                String sql = "INSERT INTO REG_CONTENT_VERSION " +
                        "(CONTENT_VERSION_ID, CONTENT_DATA) VALUES (?, ?)";

                PreparedStatement ps = conn.prepareStatement(sql);
                ps.setString(1, contentID);
                ps.setBinaryStream(2, tempFileStream, (int)contentLength);
                ps.executeUpdate();
                ps.close();

            } catch (SQLException e) {

                String msg = "Failed to write resource content to the database. " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);

            } finally {

                if (tempFileStream != null) {
                    try {
                        tempFileStream.close();

                    } catch (IOException e) {
                        String msg = "Failed to close the input stream used for writing " +
                                "the current version content to the archived content version. " +
                                e.getMessage();
                        log.error(msg, e);
                    }
                }
            }
        }

        return contentID;
    }

    public void createFileBasedContentFromDB(String contentID, Connection conn)
            throws RegistryException {

        if (FileManager.getInstance().lockFileBasedContent(contentID)) {
            // file based content is already created for this content ID. Locked it so that it will
            // not be automatically deleted.
            return;
        }

        try {
            String sql = "SELECT CONTENT_DATA FROM REG_CONTENT_VERSION WHERE CONTENT_VERSION_ID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            ResultSet result = ps.executeQuery();

            if (result.next()) {
                FileManager.getInstance().
                        createFileBasedContent(contentID, result.getBinaryStream("CONTENT_DATA"));
            }
            ps.close();

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        }
    }

    public InputStream getResourceContentStream(String contentID, Connection conn)
            throws RegistryException {

        InputStream contentStream = FileManager.getInstance().getFileBasedInputStream(contentID);
        if (contentStream != null) {
            return contentStream;
        }

        try {
            String sql = "SELECT CONTENT_DATA FROM REG_CONTENT_VERSION WHERE CONTENT_VERSION_ID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                //contentStream = getDisconnectedStream(result.getBinaryStream("CONTENT_DATA"));
                contentStream = FileManager.getInstance().
                        createFileBasedInputStream(contentID, result.getBinaryStream("CONTENT_DATA"));
            }
            ps.close();

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        }
        return contentStream;
    }

    public ResourceDO getResourceVersionDO(String resourceID, long snapshotID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT R.RID, R.VERSION, R.PATH, R.MEDIA_TYPE, R.COLLECTION, " +
                "R.CREATOR, R.CREATED_TIME, R.LAST_UPDATOR, R.LAST_UPDATED_TIME, " +
                "R.DESCRIPTION, R.CONTENT_ID, R.ASSOCIATED_SNAPSHOT_ID " +
                "FROM REG_RESOURCE_VERSION R WHERE R.RID=? AND R.VERSION=" +
                "(SELECT S.VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION S " +
                "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, snapshotID);
            ps.setString(3, resourceID);

            ResourceDO resourceVersionDO = null;
            ResultSet result = ps.executeQuery();
            if (result.next()) {
                resourceVersionDO = new ResourceDO();

                resourceVersionDO.setID(result.getString("RID"));
                resourceVersionDO.setVersion(result.getLong("VERSION"));
                resourceVersionDO.setPath(result.getString("PATH"));
                resourceVersionDO.setMediaType(result.getString("MEDIA_TYPE"));
                resourceVersionDO.setCollection(result.getInt("COLLECTION"));
                resourceVersionDO.setAuthor(result.getString("CREATOR"));
                resourceVersionDO.setCreatedOn(result.getTimestamp("CREATED_TIME"));
                resourceVersionDO.setLastUpdator(result.getString("LAST_UPDATOR"));
                resourceVersionDO.setLastUpdatedOn(result.getTimestamp("LAST_UPDATED_TIME"));
                resourceVersionDO.setDescription(result.getString("DESCRIPTION"));
                resourceVersionDO.setContentID(result.getString("CONTENT_ID"));
                resourceVersionDO.setAssociatedSnapshotID(result.getLong("ASSOCIATED_SNAPSHOT_ID"));
            }

            return resourceVersionDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + " for snapshot ID " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public List<String> getChildIDs(String resourceID, long version) throws RegistryException {

        List <String> childIDs = new ArrayList();

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT CHILD_RID FROM REG_DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet results = ps.executeQuery();
            while (results.next()) {
                childIDs.add(results.getString("CHILD_RID"));
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get child resource IDs of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return childIDs;
    }

    /**
     * Given a resource ID and a version number there may or may not be a snapshot ID. That is,
     * whenever a snapshot is created, an associating version of the resource is prepared. But,
     * when creating a snapshot of a collection, new versions will be created for all dirty child
     * resources/collections. Such resources/collection versions do not have a matching snapshot ID.
     * Such versions are just a part of a snapshot of another collection. And it is possible to
     * have more than one matching snapshot for a given resource ID and a version. A simple
     * example is creating two snapshots of a resource, without modifying the resource.
     *
     * This method returns the latest (highest) snapshot ID for the given resource ID and
     * the version. If a mathcing snapshot is not found, it will return -1.
     *
     * @param resourceID UUID of the resource
     * @param version Version number of the resource.
     * @return Matching snapshot number if available. -1 otherwise.
     * @throws RegistryException  something went wrong
     */
    public long getMatchingSnapshotID(String resourceID, long version) throws RegistryException {

        List <Long> actualSnapshotIDs = getSnapshotIDsForResourceID(resourceID);
        if (actualSnapshotIDs.size() == 0) {
            return -1;
        }

        List <Long> possibleSnapshotIDs = getSnapshotIDs(resourceID, version);

        long matchingSID = -1;
        for (int i = 0; i < actualSnapshotIDs.size(); i++) {
            long actualSID = actualSnapshotIDs.get(i);
            if (possibleSnapshotIDs.contains(actualSID)) {
                matchingSID = actualSID;
            }
        }

        return matchingSID;
    }

    /**
     * For a given snapshot, there can be at most one resource in a given path. But there may be
     * np resources for a given path or the given path may not be included in a given snapshot.
     * Therefore, given a path and snapshot ID we should get an unique resource or no resources
     * at all.
     *
     * @param path Path of the resource, at the time where snapshot is created
     * @param snapshotID Snapshot ID
     * @return Resource ID of the resource at given path in given snapshot, if a resource exists
     * at given path. Null otherwise.
     * @throws RegistryException For database related errors
     */
    public String getResourceID(String path, long snapshotID) throws RegistryException {

        List <ResourceVersionHolder> pathMatches = getResourceVersionsForPath(path);
        List <ResourceVersionHolder> sidMatches = getResourceVersionsForSnapshotID(snapshotID);

        Iterator <ResourceVersionHolder> pathMatchIterator = pathMatches.iterator();
        while (pathMatchIterator.hasNext()) {
            ResourceVersionHolder rsv = pathMatchIterator.next();
            if (sidMatches.contains(rsv)) {
                return rsv.getResourceID();
            }
        }

        return null;
    }

    public void removeSnapshotsOfResource(String resourceID) throws RegistryException {

        invalidateAssociatedSnapshot(resourceID);

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_SNAPSHOT WHERE ROOT_ID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to remove snapshots of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void invalidateAssociatedSnapshot(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "UPDATE REG_RESOURCE SET ASSOCIATED_SNAPSHOT_ID=? WHERE RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, -1);
            ps.setString(2, resourceID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to invalidate associated snapshot of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private List <ResourceVersionHolder> getResourceVersionsForPath(String path)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RID, VERSION FROM REG_RESOURCE_VERSION WHERE PATH=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, path);

            List <ResourceVersionHolder> rsvs = new ArrayList <ResourceVersionHolder> ();
            ResultSet results = ps.executeQuery();
            while(results.next()) {
                String resourceID = results.getString("RID");
                long version = results.getLong("VERSION");
                ResourceVersionHolder rsv = new ResourceVersionHolder(resourceID, version);
                rsvs.add(rsv);
            }
            ps.close();

            return rsvs;

        } catch (SQLException e) {

            String msg = "Failed to get resource ID and version for resource path " +
                    path + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private List <ResourceVersionHolder> getResourceVersionsForSnapshotID(long snapshotID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT RID, VERSION FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE SNAPSHOT_ID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);

            List <ResourceVersionHolder> rsvs = new ArrayList <ResourceVersionHolder> ();
            ResultSet results = ps.executeQuery();
            while(results.next()) {
                String resourceID = results.getString("RID");
                long version = results.getLong("VERSION");
                ResourceVersionHolder rsv = new ResourceVersionHolder(resourceID, version);
                rsvs.add(rsv);
            }
            ps.close();

            return rsvs;

        } catch (SQLException e) {

            String msg = "Failed to get resource ID and version for snapshot ID " +
                    snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private List<Long> getSnapshotIDsForResourceID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM REG_SNAPSHOT WHERE ROOT_ID=? ORDER BY SNAPSHOT_ID DESC";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet results = ps.executeQuery();
            List <Long> snapshotIDs = new ArrayList <Long> ();
            while(results.next()) {
                long snapshotNumber = results.getLong("SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }
            ps.close();

            return snapshotIDs;

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }

    private List<Long> getSnapshotIDs(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM REG_SNAPSHOT_RESOURCE_VERSION WHERE RID=? AND VERSION=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet results = ps.executeQuery();

            List <Long> sids = new ArrayList <Long> ();
            while (results.next()) {
                sids.add(results.getLong("SNAPSHOT_ID"));
            }
            ps.close();

            return sids;

        } catch (SQLException e) {

            String msg = "Failed to matching snapshot IDs for resource ID " +
                    resourceID + " and version " + version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }
}
