/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.dataobjects.PropertyDO;
import org.wso2.registry.jdbc.dataobjects.ResourceDO;
import org.wso2.registry.jdbc.utils.Transaction;

import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class RestoreUtilDAO {

    private static Log log = LogFactory.getLog(RestoreUtilDAO.class);

    private ResourceDAO resourceDAO = new ResourceDAO();
    private ResourceVersionDAO resourceVersionDAO = new ResourceVersionDAO();

    public String replaceContentVersion(ResourceDO resourceDO, ResourceDO resourceVersionDO)
            throws RegistryException {

        Connection conn = Transaction.getConnection();
        InputStream contentStream = resourceVersionDAO.
                getResourceContentStream(resourceVersionDO.getContentID(), conn);
        resourceDAO.updateContent(resourceDO.getContentID(), contentStream);

        return resourceDO.getContentID();
    }

    public void mergeResourceVersion(ResourceDO resourceDO,
                                     ResourceDO resourceVersionDO, String contentID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "UPDATE REG_RESOURCE SET EQUIVALENT_VERSION=?, MEDIA_TYPE=?, " +
                "LAST_UPDATOR=?, LAST_UPDATED_TIME=?, DESCRIPTION=?, CONTENT_ID=?, " +
                "ASSOCIATED_SNAPSHOT_ID=? WHERE RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            ps.setLong(1, resourceVersionDO.getVersion());
            ps.setString(2, resourceVersionDO.getMediaType());
            ps.setString(3, resourceVersionDO.getLastUpdator());
            ps.setTimestamp(4, resourceVersionDO.getLastUpdatedOn());
            ps.setString(5, resourceVersionDO.getDescription());
            ps.setString(6, contentID);
            ps.setLong(7, resourceVersionDO.getAssociatedSnapshotID());
            ps.setString(8, resourceDO.getID());

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to merge the revision " + resourceVersionDO.getVersion() +
                    " with the current version of resource " + resourceDO.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void replaceProperties(String resourceID, long versionNumber) throws RegistryException {

        deleteProperties(resourceID);
        List <PropertyDO> properties = getProperties(resourceID, versionNumber);
        setProperties(resourceID, properties);
    }

    private void setProperties(String resourceID, List<PropertyDO> properties)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_PROPERTY (RID, NAME, PROPERTY_VALUE) VALUES (?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            Iterator <PropertyDO> iProps = properties.iterator();
            while (iProps.hasNext()) {
                PropertyDO propertyDO = iProps.next();
                ps.setString(1, resourceID);
                ps.setString(2, propertyDO.getName());
                ps.setString(3, propertyDO.getValue());
                ps.executeUpdate();
                ps.clearParameters();
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to set properties for resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private List<PropertyDO> getProperties(String resourceID, long versionNumber)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT NAME, PROPERTY_VALUE FROM REG_PROPERTY_VERSION WHERE RID=? AND VERSION=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, versionNumber);

            ResultSet result = ps.executeQuery();

            List <PropertyDO> properties = new ArrayList();
            while (result.next()) {
                PropertyDO propertyDO = new PropertyDO();
                propertyDO.setName(result.getString("NAME"));
                propertyDO.setValue(result.getString("PROPERTY_VALUE"));

                properties.add(propertyDO);
            }
            ps.close();

            return properties;

        } catch (SQLException e) {

            String msg = "Failed to read properties revision " + versionNumber +
                    " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void deleteProperties(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_PROPERTY WHERE RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to delete properties of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public String copyContentVersion(ResourceDO resourceVersionDO)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        InputStream contentStream = resourceVersionDAO.
                getResourceContentStream(resourceVersionDO.getContentID(), conn);
        String contentID = resourceDAO.addContent(contentStream);

        return contentID;
    }

    public void copyResourceVersion(
            ResourceDO resourceVersionDO, String contentID, String resourcePath)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_RESOURCE (RID, PATH, MEDIA_TYPE, COLLECTION, CREATOR, " +
                "CREATED_TIME, LAST_UPDATOR, LAST_UPDATED_TIME, DESCRIPTION, CONTENT_ID, " +
                "EQUIVALENT_VERSION, ASSOCIATED_SNAPSHOT_ID) " +
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            ps.setString(1, resourceVersionDO.getID());
            ps.setString(2, resourcePath);
            ps.setString(3, resourceVersionDO.getMediaType());
            ps.setInt(4, resourceVersionDO.getCollection());
            ps.setString(5, resourceVersionDO.getAuthor());
            ps.setTimestamp(6, resourceVersionDO.getCreatedOn());
            ps.setString(7, resourceVersionDO.getLastUpdator());
            ps.setTimestamp(8, resourceVersionDO.getLastUpdatedOn());
            ps.setString(9, resourceVersionDO.getDescription());
            ps.setString(10, contentID);
            ps.setLong(11, resourceVersionDO.getVersion());
            ps.setLong(12, resourceVersionDO.getAssociatedSnapshotID());

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to copy the revision " + resourceVersionDO.getVersion() +
                    " to the current version of resource " + resourceVersionDO.getPath() + ". " +
                    e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void copyProperties(String resourceID, long versionNumber) throws RegistryException {

        List <PropertyDO> properties = getProperties(resourceID, versionNumber);
        setProperties(resourceID, properties);
    }

    public void replaceDependencies(String resourceID, List<String> versionedChildIDs) 
            throws RegistryException {

        deleteDependencies(resourceID);
        addDependencies(resourceID, versionedChildIDs);
    }

    private void addDependencies(String resourceID, List<String> versionedChildIDs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO REG_DEPENDENCY (PARENT_RID, CHILD_RID) VALUES (? ,?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            Iterator <String> iChildIDs = versionedChildIDs.iterator();
            while (iChildIDs.hasNext()) {
                String childID = iChildIDs.next();

                ps.setString(1, resourceID);
                ps.setString(2, childID);
                ps.executeUpdate();
                ps.clearParameters();
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add dependencies to resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private void deleteDependencies(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_DEPENDENCY WHERE PARENT_RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {
            String msg = "Failed to delete dependencies of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }

    public void deleteContent(String contentID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "DELETE FROM REG_CONTENT WHERE CONTENT_ID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);
            ps.executeUpdate();

        } catch (SQLException e) {

            String msg = "Failed to delete content with content ID " +
                    contentID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }
}
