/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.handlers;

import org.wso2.registry.Resource;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.dao.ResourceDAO;
import org.wso2.registry.utils.AuthorizationUtils;

/**
 * Base class of all handler implementations. Provides the methods that handlers should implement.
 * This class also provides the datasource, user realm, registry and repository instances to be
 * used by handler implementations.
 *
 * Handlers can be chained by providing necessary filter combinations. But in such cases, handler
 * authors should make sure that handlers in the chain do not perform conflicting operations.
 * Unless there is a critical requirement and handler authors are confident that handlers do not
 * have negative impact on each other, it is recommended that handlers are configured to execute
 * only one handler per request.
 *
 * Handler instances may be accessed concurrently by multiple threads. Therefore, handlers should
 * be thread safe. It is recommended that handlers are made stateless, instead of synchronizing
 * them as it could become a performance bottleneck in highly concurrent environments.
 *
 * Implementations of handlers should be optimized to take the minimum time for processing. As the
 * handlers are executed are always executed before executing the generic database layer code,
 * time consuming operations in handlers could slow down the whole registry.
 */
public abstract class Handler {
    
    /** ResourceDAO for directly accessing resources. */
    protected ResourceDAO resourceDAO = new ResourceDAO();

    /** Some common authorization tasks are implemented in this util. */
    protected AuthorizationUtils authorizationUtil = new AuthorizationUtils();

    /**
     * Processes the GET action for resource path of the requestContext.
     *
     * @param requestContext Information about the current request.
     *
     *requestContext.resourcePath: Path of the resource
     *
     *requestContext.resource: Resource at the given path. This can be null if no other handler has
     * retrieved that resource so far. If it contains a value, matching handlers are free to do any
     * change to the resource, even thay can replace the resource with completely new instance.
     *
     * @return Resource instance if the handler processed the GET action successfully.
     *
     * @throws org.wso2.registry.exceptions.RegistryException If the media type handler is supposed to handle
     *  the get on the media type and if the get fails due a handler specific error
     */
    public Resource get(RequestContext requestContext) throws RegistryException {
        return null;
    }

    /**
     * Processes the PUT action. Actual path to which the resource is put may differ from the path
     * given in the requestContext.resourcePath. Therefore, after putting the resource, the actual
     * path to which the resource is put is set in the requestContext.actualPath.
     *
     * @param requestContext Information about the current request.
     *
     * requestContext.resourcePath: Path to put the resource.
     * requestContext.resource: Resource to put
     *
     * @throws RegistryException If the media type handler is supposed to handle the put on the
     * media type and if the put fails due a handler specific error
     */
    public void put(RequestContext requestContext) throws RegistryException {}

    /**
     * Creates a resource in the given path by fetching the resource content from the given URL.
     *
     * @param requestContext Information about the current request.
     *
     * requestContext.resourcePath: Path to add the new resource.
     *
     * requestContext.sourceURL: URL to fetch the resource content
     *
     * requestContext.resource: Resource instance containing the metadata for the resource to be
     * imported. Once import is done, new resource is created combining the metadata of this
     * metadata object and the imported content.
     *
     * @throws RegistryException If the media type handler is supposed to handle the import on the
     * media type and if the import fails due a handler specific error
     */
    public void importResource(RequestContext requestContext) throws RegistryException {}

    /**
     * Processes the DELETE action of the media type.
     *
     * @param requestContext Information about the current request.
     *
     * requestContext.resourcePath: path path of the resource to be deleted.
     *
     * @throws RegistryException If the media type handler is supposed to handle the delete on the
     * media type and if the delete fails due a handler specific error
     */
    public void delete(RequestContext requestContext) throws RegistryException {}

    /**
     * Invokes when a child resource is added. Only the media type handlers of collection resources
     * may have a meaningfull implementation of this method.
     *
     * @param requestContext
     *          requestContext.resourcePath: path of the parent collection
     *          requestContext.resource: New child resource to be added
     * 
     * @throws RegistryException If the media type handler is supposed to handle the putChild on the
     * media type and if the putChild fails due a handler specific error
     */
    public void putChild(RequestContext requestContext) throws RegistryException {}

    /**
     * Invokes when a child resource is imported. Only the media type handlers of collection
     * resources may have a meaningfull implementation of this method.
     *
     * @param requestContext
     *          requestContext.resourcePath
     *
     * @throws RegistryException If the media type handler is supposed to handle the importChild
     * on the media type and if the importChild fails due a handler specific error
     */
    public void importChild(RequestContext requestContext) throws RegistryException {}

}
