/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.handlers;

import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.Resource;
import org.wso2.registry.utils.RegistryUtils;
import org.wso2.registry.jdbc.handlers.filters.Filter;

import java.util.*;

/**
 * Manages the handlers and thier invocations. Handlers, which are required to be invoked must be
 * registered with this class using the addHanlder(...) method. Handlers are allowed for a subset
 * of Registry API methods (see Java doc of Handler class for more detrails). Handler authors can
 * further specify exactly for which methods thier handlers should engage. For example, handler
 * author may specify his handler should be engaged only for GET and DELETE operations. This class
 * maintains separate maps for each supported operation and puts handlers to corresponding maps
 * once registered.
 *
 * Handlers are registered with a Filter instance. Filter decides (based on the request) whether or
 * not to invoke its associated handler.
 *
 * This class has a separate method for each supported operation. These methods are invoked by
 * BasicRegistry once such method is invoked in the Registry API. Then this class iterates through
 * the map corresponding to the invoked method and evalautes filters of registered handlers. For
 * each filter which evaluates to true, its associated handler will be invoked. This process
 * continues either till there is no more handlers or till the processingComplete parameter of
 * the RequestContext is set to true.
 *
 * There is only one instance of this class exists per registry instance.
 */
public class HandlerManager {

    /**
     * Maps of <Filter, Handler> pairs. A map exists for each supported operation.
     */
    private Map <Filter, Handler> getHandlerMap = new HashMap <Filter, Handler> ();
    private Map <Filter, Handler> putHandlerMap = new HashMap <Filter, Handler> ();
    private Map <Filter, Handler> deleteHandlerMap = new HashMap <Filter, Handler> ();
    private Map <Filter, Handler> importHandlerMap = new HashMap <Filter, Handler> ();
    private Map <Filter, Handler> putChildHandlerMap = new HashMap <Filter, Handler> ();
    private Map <Filter, Handler> importChildHandlerMap = new HashMap <Filter, Handler> ();

    /**
     * Registers handlers with the handler manager. Each handler should be registered with a
     * Filter. If a handler should be engaged only to a subset of allowed methods, those methods
     * can be specified as a string array.
     *
     * @param methods Methods for which the registered handler should be engaged. Allowed values in
     * the string array are "GET", "PUT", "IMPORT", "DELETE", "PUT_CHILD", and "IMPORT_CHILD".
     * If null is given, handler will be engaged to all methods.
     *
     * @param filter Filter instance associated with the handler.
     *
     * @param handler Handler instance to be registered.
     */
    public void addHandler(String[] methods, Filter filter, Handler handler) {

        if (methods == null || RegistryUtils.containsString(Filter.GET, methods)) {
            getHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT, methods)) {
            putHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.DELETE, methods)) {
            deleteHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT, methods)) {
            importHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.PUT_CHILD, methods)) {
            putChildHandlerMap.put(filter, handler);
        }

        if (methods == null || RegistryUtils.containsString(Filter.IMPORT_CHILD, methods)) {
            importChildHandlerMap.put(filter, handler);
        }
    }

    /**
     * Manages the handler invocations of GET method.
     *
     * @param requestContext Details of the request.
     * @return Resource requested from the GET operation.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public Resource get(RequestContext requestContext) throws RegistryException {

        Resource resource = null;

        Iterator<Filter> filters = getHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleGet(requestContext)) {
                Handler handler = getHandlerMap.get(filter);
                resource = handler.get(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        return resource;
    }

    /**
     * Manages the handler invocations of PUT method.
     *
     * @param requestContext Details of the request.
     * @return Path where the resource is actually stored. Subsequent accesses to the resource
     * should use this path.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String put(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = putHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handlePut(requestContext)) {
                putHandlerMap.get(filter).put(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        if (requestContext.isProcessingComplete()) {
            return requestContext.getActualPath();
        } else {
            return null;
        }
    }

    /**
     * Manages the handler invocations of IMPORT method.
     *
     * @param requestContext Details of the request.
     * @return Path where the resource is actually stored. Subsequent accesses to the resource
     * should use this path.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public String importResource(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = importHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleImportResource(requestContext)) {
                importHandlerMap.get(filter).importResource(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }

        if (requestContext.isProcessingComplete()) {
            return requestContext.getActualPath();
        } else {
            return null;
        }
    }

    /**
     * Manages the handler invocations of DELETE method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void delete(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = deleteHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleDelete(requestContext)) {
                deleteHandlerMap.get(filter).delete(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of PUT_CHILD method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void putChild(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = putChildHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handlePutChild(requestContext)) {
                putChildHandlerMap.get(filter).putChild(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }

    /**
     * Manages the handler invocations of IMPORT_CHILD method.
     *
     * @param requestContext Details of the request.
     * @throws RegistryException This exception is thrown for all exceptions occured inside
     * handlers or filters.
     */
    public void importChild(RequestContext requestContext) throws RegistryException {

        Iterator<Filter> filters = importChildHandlerMap.keySet().iterator();
        while (filters.hasNext()) {
            Filter filter = filters.next();
            if (filter.handleImportChild(requestContext)) {
                importChildHandlerMap.get(filter).importChild(requestContext);
            }

            if (requestContext.isProcessingComplete()) {
                break;
            }
        }
    }
}
