/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.handlers.filters;

import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.jdbc.Repository;
import org.wso2.registry.jdbc.handlers.RequestContext;

/**
 * Base class of all filter implementations. All handlers have to be registered in the jdbc registry
 * with a filter implementation. Filter implementations determine the conditions to invoke the
 * associating handler.
 */
public abstract class Filter {

    /**
     * Supported method names 
     */
    public static final String GET = "GET";
    public static final String PUT = "PUT";
    public static final String DELETE = "DELETE";
    public static final String IMPORT = "IMPORT";
    public static final String PUT_CHILD = "PUT_CHILD";
    public static final String IMPORT_CHILD = "IMPORT_CHILD";

    /**
     * Determines whether the associating handler should handle the get action.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     *
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handleGet(RequestContext requestContext) throws RegistryException;

    /**
     * Determines whether the associating handler should handle the put action.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     *
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handlePut(RequestContext requestContext) throws RegistryException;

    /**
     * Determines whether the associating handler should handle the import resource action.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     *
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handleImportResource(RequestContext requestContext)
            throws RegistryException;

    /**
     * Determines whether the associating handler should handle the delete action.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     *
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handleDelete(RequestContext requestContext) throws RegistryException;

    /**
     * Determines whether the associating handler should handle the putChild action. putChild action
     * occurs when it is attempted to put a resource as a child of the resource referred by
     * requestContext.parentPath or to the resource requestContext.parentCollection.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     *
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handlePutChild(RequestContext requestContext) throws RegistryException;

    /**
     * Determines whether the associating handler should handle the importChild action.
     *
     * @param requestContext Information about the current requestContext.
     *
     * @return true if the associating handler should handle the request. false otherwise.
     * 
     * @throws RegistryException Filter impls should deal with the specific exceptions and throw
     * a RegistryException if the exception has to be propagated to the surface.
     */
    public abstract boolean handleImportChild(RequestContext requestContext)
            throws RegistryException;
}
