/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.servlet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.*;
import org.wso2.registry.config.HandlerConfiguration;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.jdbc.EmbeddedRegistry;
import org.wso2.registry.servlet.utils.Utils;
import org.wso2.registry.session.UserRegistry;
import org.wso2.registry.users.AccessControlAdmin;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.accesscontrol.AccessControlConstants;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URLDecoder;
import java.util.Iterator;

/** Servlet for providing REST API for the registry. */
public class RegistryServlet extends HttpServlet {

    private static final Log log = LogFactory.getLog(RegistryServlet.class);
    protected transient ServletConfig servletConfig;

    //To store the context path
    private String contextRoot = null;

    public void init(final ServletConfig config) throws ServletException {
        super.init(config);
        EmbeddedRegistry embeddedRegistry;

        try {
            // read the registry.xml file from the configured location. if not configured, read the
            // default registry.xml file bundled with the webapp.
            String configPath = config.getInitParameter(RegistryConstants.REGISTRY_CONFIG_PATH);
            InputStream configStream;
            if (configPath != null) {
                try {
                    configStream = new FileInputStream(configPath);
                } catch (FileNotFoundException e) {
                    throw new ServletException("Couldn't find specified config file '" +
                                               configPath + "'", e);
                }
            } else {
                configStream =
                        config.getServletContext().getResourceAsStream("/WEB-INF/registry.xml");
            }

            RegistryContext registryContext = new RegistryContext(configStream,
                    new RegistryContext.RegURLSupplier() {
                        public String getURL() {
                            return config.getServletContext().getRealPath("/WEB-INF");
                        }
                    });
            RegistryContext.setSingleton(registryContext);

            embeddedRegistry = registryContext.getEmbeddedRegistry();

            // create a system registry and put it in the context
            UserRegistry systemRegistry = embeddedRegistry.getUserRegistry(
                    RegistryConstants.SYSTEM_USER, RegistryConstants.SYSTEM_PASSWORD);

            // add configured handers to the jdbc registry
            // note: no need to do this here. this is done inside the registry context
            //Iterator<HandlerConfiguration> handlers =
            //        registryContext.getHandlerConfigurations().iterator();
            //while (handlers.hasNext()) {
            //    HandlerConfiguration handlerConfiguration = handlers.next();
            //    registryContext.getHandlerManager().addHandler(0,
            //            handlerConfiguration.getFilter(), handlerConfiguration.getHandler());
            //}

            // create system resources
            if (!systemRegistry.resourceExists("/system")) {

                Collection systemCollection = systemRegistry.newCollection();
                String systemDesc = "This collection is used to store system data of the " +
                        "WSO2 Registry server. User nor the admins of the registry are not expected " +
                        "to edit any content of this collection. Changing content of this collection " +
                        "may result in unexpected behaviors.";
                systemCollection.setDescription(systemDesc);
                systemRegistry.put("/system", systemCollection);

                Collection advancedQueryCollection = systemRegistry.newCollection();
                String advaceDesc = "This collection is used to store auto generated queries " +
                        "to support various combinations of advanced search criteria. " +
                        "This is initialy empty and gets filled as advanced search is " +
                        "executed from the web UI.";
                advancedQueryCollection.setDescription(advaceDesc);
                systemRegistry.put("/system/queries/advanced", advancedQueryCollection);
            }

            try {
                AccessControlAdmin ac = systemRegistry.getUserRealm().getAccessControlAdmin();
                ac.clearResourceAuthorizations("/system");

                ac.authorizeUser(RegistryConstants.SYSTEM_USER, "/system", ActionConstants.GET);
                ac.authorizeUser(RegistryConstants.SYSTEM_USER, "/system", ActionConstants.PUT);
                ac.authorizeUser(RegistryConstants.SYSTEM_USER, "/system", ActionConstants.DELETE);
                ac.authorizeUser(
                        RegistryConstants.SYSTEM_USER, "/system", AccessControlConstants.AUTHORIZE);

                ac.authorizeUser(RegistryConstants.ADMIN_USER, "/system", ActionConstants.GET);

                ac.authorizeRole(RegistryConstants.ADMIN_ROLE, "/system", ActionConstants.GET);

                // any user should be able to execute auto generated queries, though the results
                // of such queries are filtered to match current users permission level.
                ac.authorizeRole(RegistryConstants.EVERYONE_ROLE,
                        "/system/queries/advanced", ActionConstants.GET);

            } catch (UserStoreException e) {
                String msg = "Failed to set permissions for the system collection.";
                log.fatal(msg, e);
                throw new RegistryException(msg, e);
            }

            // todo: we should make this decision according to the registry.xml or web.xml config
            //try {
            //    RemoteRegistry remote = new RemoteRegistry(new URL("http://localhost:8080/wso2registry/atom"));
            //    config.getServletContext().setAttribute(RegistryConstants.REGISTRY, remote);
            //} catch (MalformedURLException e) {
            //    config.getServletContext().setAttribute(RegistryConstants.REGISTRY, coreRegistry);
            //    e.printStackTrace();
            //}

            config.getServletContext().setAttribute(RegistryConstants.REGISTRY, embeddedRegistry);

            //config.getServletContext()
            //        .setAttribute(RegistryConstants.REGISTRY_REALM, registryRealm);
            System.getProperties().put(RegistryConstants.REGISTRY, embeddedRegistry);
            System.getProperties().put(RegistryConstants.SYSTEM_REGISTRY, systemRegistry);
            //System.getProperties().put(RegistryConstants.REGISTRY_REALM, registryRealm);

        } catch (RegistryException e) {
            String msg = "Registry initialization failed. " + e.getMessage();
            log.fatal(msg, e);
            throw new ServletException(msg, e);
        }

        log.info(Messages.getMessage("server.initalized"));
    }

    public void init() throws ServletException {
        if (servletConfig != null) {
            init(servletConfig);
        }
    }

    protected void doPost(HttpServletRequest httpServletRequest,
                          HttpServletResponse httpServletResponse)
            throws ServletException, IOException {
        initContextRoot(httpServletRequest);
        String uri = httpServletRequest.getRequestURI();
        String p1 = uri.substring(contextRoot.length(), uri.length());
        if (p1.equals("/web/addResource")) {
            handleFileUpload(httpServletRequest, httpServletResponse);
        }
    }

    private void handleFileUpload(HttpServletRequest request, HttpServletResponse response)
            throws IOException, ServletException {
        try {
            Registry registry = Utils.getSecureRegistry(request);
            ResourceImpl fileElement =
                    FileUploadUtil.processUpload(request);
            String path = fileElement.getPath();
            registry.put(path, fileElement);
            request.getSession().setAttribute(RegistryConstants.STATUS_MESSAGE_NAME,
                    "Resource " + path +
                            " was successfully added to the registry.");
            response.setContentType("text/html");
            request.getRequestDispatcher(RegistryConstants.RESOURCES_JSP)
                    .forward(request, response);

        } catch (RegistryException e) {

            e.printStackTrace();
            request.getSession().setAttribute(RegistryConstants.STATUS_MESSAGE_NAME,
                    e.getMessage());
            response.setContentType("text/html");
            request.getRequestDispatcher(RegistryConstants.RESOURCES_JSP)
                    .forward(request, response);
        }
    }

    protected void doGet(HttpServletRequest request,
                         HttpServletResponse response)
            throws ServletException, IOException {
        initContextRoot(request);

        String uri = URLDecoder.decode(request.getRequestURI(), "UTF-8");

        String controlPart = uri.substring(contextRoot.length());

        if (controlPart.startsWith(
                RegistryConstants.PATH_SEPARATOR + RegistryConstants.RESOURCES_PATH)) {

            String path;

            if (uri.equals("") || uri.endsWith(RegistryConstants.RESOURCES_PATH +
                    RegistryConstants.PATH_SEPARATOR)) {
                path = RegistryConstants.ROOT_PATH;
            } else {
                path = uri.substring((contextRoot + RegistryConstants.PATH_SEPARATOR +
                        RegistryConstants.RESOURCES_PATH).length(), uri.length());
            }

            // if user is browsing an old version of the resource, we append it to the path, so that
            // the backend registry gives the details of the version
            //String qPart = request.getQueryString();
            //if (qPart != null && qPart.startsWith("v")) {
            //    path = path + "?" + qPart;
            //}

            ResourceRequestProcessor.processResourceGET(request, response, path);
        }
    }

    /**
     * To find the name of the war , so getting that from the request context path
     *
     * @param req :  HttpServletRequest
     */
    public void initContextRoot(HttpServletRequest req) {
        if (contextRoot != null && contextRoot.trim().length() != 0) {
            return;
        }
        String contextPath = req.getContextPath();
        //handling ROOT scenario, for servlets in the default (root) context, this method returns ""
        if (contextPath != null && contextPath.length() == 0) {
            contextPath = "/";
        }
        this.contextRoot = contextPath;
    }

    protected void renderView(String jspName,
                              HttpServletRequest httpServletRequest,
                              HttpServletResponse httpServletResponse)
            throws IOException, ServletException {
        httpServletResponse.setContentType("text/html");
        httpServletRequest.getRequestDispatcher(jspName)
                .include(httpServletRequest, httpServletResponse);
    }

//    private void setErrorMessage(HttpServletRequest request, String message) {
//        request.getSession().setAttribute(RegistryConstants.ERROR_MESSAGE, message);
//    }
}
