/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.session;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.authenticator.Authenticator;
import org.wso2.authenticator.AuthenticatorException;
import org.wso2.registry.*;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.jdbc.handlers.Handler;
import org.wso2.registry.jdbc.handlers.filters.Filter;
import org.wso2.registry.jdbc.realm.pathbased.PathBasedRegistryRealm;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.accesscontrol.AuthorizingRealm;
import org.wso2.registry.users.accesscontrol.AuthorizingRealmConfig;

import javax.sql.DataSource;
import java.util.Date;
import java.util.Map;

/**
 * Maintains information required for particular user. When registry functionality is invoked
 * through this objects of this class, user's information is passes along the thread of
 * invocation. Therefore, all the actions are recorded against that user's name and all
 * authorizations are checked using that user's realm. Note that, although this class maintains
 * user details, it does not perform any authorization. Underlying registry or other entities
 * in the thread of invocation should perform authorizations as needed using the realm given by
 * this class.
 */
public class UserRegistry implements Registry {

    public static final Log log = LogFactory.getLog(UserRegistry.class);

    /**
     * User name of the user. All actions done using this registry will be recorded against this
     * user name.
     */
    private String userName;

    /**
     * Realm of the user. All actions done using this registry will be authorized using this realm.
     */
    private UserRealm userRealm;

    /**
     * Path based registry realm to be used by API users.
     */
    private PathBasedRegistryRealm pathBasedRegistryRealm;

    /**
     * Core registry to delegate the requests. All UserRegistry instances share this core
     * registry.
     */
    private Registry coreRegistry;

    public UserRegistry (String userName, String password, Registry coreRegistry,
                         UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        try {
            Authenticator authenticator = defaultRealm.getAuthenticator();

            if (!authenticator.authenticate(userName, password)) {
                String msg = "Attempted to authenticate invalid user.";
                log.info(msg);
                throw new RegistryException(msg);
            }

        } catch (AuthenticatorException e) {
            String msg = "Could not check authentication. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } catch (UserStoreException e){
            String msg = "Could not obtain the authenticator for authenticating the user " +
                    userName + ". \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        init(userName, coreRegistry, defaultRealm, dataSource);
    }

    public UserRegistry (
            String userName, Registry coreRegistry, UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        init(userName, coreRegistry, defaultRealm, dataSource);
    }

    private void init(
            String userName, Registry coreRegistry, UserRealm defaultRealm, DataSource dataSource)
            throws RegistryException {

        this.userName = userName;

        this.coreRegistry = coreRegistry;

        UserRealm userRealm;

        try {
            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userName);
            config.setCurrentUserReadable(true);
            config.setCurrentUserEditable(true);
            userRealm.init(config);
            this.userRealm = userRealm;

        } catch (UserStoreException e){
            String msg = "Could not create the realm for user " + userName + ". \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        pathBasedRegistryRealm = new PathBasedRegistryRealm(userRealm, dataSource);
    }

    public void beginTransaction() throws RegistryException {
        coreRegistry.beginTransaction();
    }

    public void commitTransaction() throws RegistryException {
        coreRegistry.commitTransaction();
    }

    public void rollbackTransaction() throws RegistryException {
        coreRegistry.rollbackTransaction();
    }

    public Resource newResource() throws RegistryException {
        return coreRegistry.newResource();
    }

    public Collection newCollection() throws RegistryException {
        return coreRegistry.newCollection();
    }

    public RegistryContext getRegistryContext() {
        return coreRegistry.getRegistryContext();
    }

    public String getUserName() {
        return userName;
    }

    /**
     * Registry API users access the user realm using this method. Registry authorizations are
     * stored based on resource UUIDs. But UUIDs are meaningless to the API users. API users
     * are expected to deal only with resource paths. Therefore, this method, returns a user realm,
     * whcih converts paths to resource UUIDs before, any resource related operation.
     *
     * @return Resource path based registry realm.
     */
    public UserRealm getUserRealm() {
        return pathBasedRegistryRealm;
    }

    public Resource get(String path) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.get(path);
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.get(path, start, pageSize);
    }

    public boolean resourceExists(String path) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.resourceExists(path);
    }

    public String put(String suggestedPath, Resource resource) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.put(suggestedPath, resource);
    }

    public void delete(String path) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.delete(path);
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata)
            throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.importResource(suggestedPath, sourceURL, metadata);
    }

    public String rename(String currentPath, String newPath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.rename(currentPath, newPath);
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.move(currentPath, newPath);
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.copy(sourcePath, targetPath);
    }

    public void createVersion(String path) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.createVersion(path);
    }

    public String[] getVersions(String path) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getVersions(path);
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.restoreVersion(versionPath);
    }

    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String sourcePath, String targetPath, String associationType)
            throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.addAssociation(sourcePath, targetPath, associationType);
    }

    public void removeAssociation(String sourcePath, String associationPaths,
                                  String associationType)
            throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.removeAssociation(sourcePath, associationPaths, associationType);
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getAllAssociations(resourcePath);
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getAssociations(resourcePath, associationType);
    }


    public void applyTag(String resourcePath, String tag) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.applyTag(resourcePath, tag);
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getResourcePathsWithTag(tag);
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getTags(resourcePath);
    }

    public void removeTag(String path, String tag) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.removeTag(path, tag);
    }

    public String addComment(String resourcePath, Comment comment) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.addComment(resourcePath, comment);
    }

    public void editComment(String commentPath, String text) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.editComment(commentPath, text);
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getComments(resourcePath);
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.rateResource(resourcePath, rating);
    }

    public float getAverageRating(String resourcePath) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getAverageRating(resourcePath);
    }

    public int getRating(String path, String userName) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getRating(path, userName);
    }

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.executeQuery(path, parameters);
    }

    public LogEntry[] getLogs(
            String resourcePath,
            int action,
            String userName,
            Date from,
            Date to,
            boolean recentFirst)
            throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getLogs(resourcePath, action, userName, from, to, recentFirst);
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);
        return coreRegistry.getLogCollection(resourcePath, action, userName, from, to, recentFirst);
    }

    public String[] getAvailableAspects() {
        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getAvailableAspects();
    }

    public void associateAspect(String resourcePath, String aspect) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.associateAspect(resourcePath, aspect);
    }

    public void invokeAspect(String resourcePath, String aspectName, String action) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        coreRegistry.invokeAspect(resourcePath, aspectName, action);
    }

    public String[] getAspectActions(String resourcePath, String aspectName) throws RegistryException {

        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.getAspectActions(resourcePath, aspectName);
    }

    public Collection searchContent(String keywords) throws RegistryException{
        CurrentSession.setUser(userName);
        CurrentSession.setRealm(userRealm);

        return coreRegistry.searchContent(keywords);
    }
}
