/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.registry.users.def;

import org.apache.axiom.om.util.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.users.UserStoreAdmin;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.def.util.DefaultDatabaseUtil;

import javax.sql.DataSource;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Manipulates data in the User Store
 * 
 * @see org.wso2.usermanager.UserStoreAdmin
 */
public class DefaultUserStoreAdmin
        extends DefaultUserStoreReader implements UserStoreAdmin {

    private static Log log = LogFactory.getLog(DefaultUserStoreAdmin.class);

    public DefaultUserStoreAdmin(DataSource dataSource) {
        super(dataSource);
    }

    public void addUser(String userName, Object credential) throws UserStoreException {

        if (userName == null || credential == null) {
            throw new UserStoreException(Messages.getMessage("nullData"));
        }

        if (!(credential instanceof String)) {
            throw new UserStoreException(Messages.getMessage("credentialTypeNotSupport"));
        }

        if (userName.trim().length() == 0) {
            throw new UserStoreException(Messages.getMessage("nullUser"));
        }

        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement addUserStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.ADD_USER_SQL);

            addUserStmt.setString(1, userName);
            MessageDigest dgst = MessageDigest.getInstance("MD5");
            dgst.update(((String) credential).getBytes());
            addUserStmt.setString(2, Base64.encode(dgst.digest()));
            addUserStmt.executeUpdate();
            dbConnection.commit();
            addUserStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } catch (NoSuchAlgorithmException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorCreatingPasswordDigest"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }

    }

    public void updateUser(String userName, Object newCredential, Object oldCredential)
            throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement getUserStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.GET_USER_SQL);
            getUserStmt.setString(1, userName);

            ResultSet rs = getUserStmt.executeQuery();
            if (rs.next()) {
                if (oldCredential != null) {
                    MessageDigest dgst = MessageDigest.getInstance("MD5");
                    dgst.update(((String) oldCredential).getBytes());
                    String dbCred = rs.getString(DefaultRealmConstants.COLUMN_NAME_CREDENTIAL);
                    if (!Base64.encode(dgst.digest()).equals(dbCred)) {
                        throw new UserStoreException(Messages.getMessage("oldPasswordIncorrect"));
                    }
                }
            } else {
                throw new UserStoreException(Messages.getMessage("oldPasswordIncorrect"));
            }

            PreparedStatement updateUserStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.UPDATE_USER_SQL);
            MessageDigest dgst = MessageDigest.getInstance("MD5");
            dgst.update(((String) newCredential).getBytes());
            updateUserStmt.setString(1, Base64.encode(dgst.digest()));
            updateUserStmt.setString(2, userName);
            updateUserStmt.executeUpdate();
            dbConnection.commit();
            updateUserStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } catch (NoSuchAlgorithmException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorCreatingPasswordDigest"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void updateUser(String userName, Object newCredential) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement updateUserStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.UPDATE_USER_SQL);
            MessageDigest dgst = MessageDigest.getInstance("MD5");
            dgst.update(((String) newCredential).getBytes());
            updateUserStmt.setString(1, Base64.encode(dgst.digest()));
            updateUserStmt.setString(2, userName);
            updateUserStmt.executeUpdate();
            dbConnection.commit();
            updateUserStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } catch (NoSuchAlgorithmException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorCreatingPasswordDigest"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void deleteUser(String userName) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement deleteUserStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.DELETE_USER_SQL);
            deleteUserStmt.setString(1, userName);
            deleteUserStmt.executeUpdate();
            dbConnection.commit();
            deleteUserStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void setUserProperties(String userName, Map properties) throws UserStoreException {

        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);

            int userid = DefaultDatabaseUtil.getUserId(dbConnection, userName);
            if (userid == -1) {
                throw new UserStoreException(Messages.getMessage("nullUser"));
            }
            
            
            PreparedStatement deleteUserPropsStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.DELETE_USER_ATTRIBUTE_SQL);
            deleteUserPropsStmt.setInt(1, userid);
            deleteUserPropsStmt.executeUpdate();

            PreparedStatement setUserPropertiesStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.ADD_USER_ATTRIBUTE_SQL);

            if (properties != null) {
                Iterator ite = properties.entrySet().iterator();
                while (ite.hasNext()) {
                    Entry entry = (Entry) ite.next();
                    String key = (String) entry.getKey();
                    String value = (String) entry.getValue();
                    
                    if(key.contains("%")){
                        throw new UserStoreException(Messages.getMessage("errorSpecialCharactor",
                                new String[]{"%"}));
                    }
                    
                    if (value != null) {
                        setUserPropertiesStmt.setString(1, key);
                        setUserPropertiesStmt.setString(2, value);
                        setUserPropertiesStmt.setInt(3, userid);
                        setUserPropertiesStmt.executeUpdate();
                    }
                }
            }
            dbConnection.commit();

        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }

    }

    public void setRoleProperties(String roleName, Map properties) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);

            int roleid = DefaultDatabaseUtil.getRoleId(dbConnection, roleName);
            if (roleid == -1) {
                throw new UserStoreException(Messages.getMessage("nullRole"));
            }

            PreparedStatement deleteRolePropsStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.DELETE_ROLE_ATTRIBUTE_SQL);
            deleteRolePropsStmt.setInt(1, roleid);
            deleteRolePropsStmt.executeUpdate();

            PreparedStatement setRolePropertiesStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.ADD_ROLE_ATTRIBUTE_SQL);

            Iterator ite = properties.entrySet().iterator();
            while (ite.hasNext()) {
                Entry entry = (Entry) ite.next();
                String key = (String) entry.getKey();
                String value = (String) entry.getValue();
                if (value != null) {
                    setRolePropertiesStmt.setString(1, key);
                    setRolePropertiesStmt.setString(2, value);
                    setRolePropertiesStmt.setInt(3, roleid);
                    setRolePropertiesStmt.executeUpdate();
                }
            }
            dbConnection.commit();

            deleteRolePropsStmt.close();
            setRolePropertiesStmt.close();

        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void addRole(String roleName) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement addRoleStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.ADD_ROLE_SQL);
            addRoleStmt.setString(1, roleName);
            addRoleStmt.execute();
            dbConnection.commit();
            addRoleStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void deleteRole(String roleName) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement deleteRoleStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.DELETE_ROLE_SQL);
            deleteRoleStmt.setString(1, roleName);
            deleteRoleStmt.execute();
            dbConnection.commit();
            deleteRoleStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void addUserToRole(String userName, String roleName) throws UserStoreException {

        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);

            int userid = DefaultDatabaseUtil.getUserId(dbConnection, userName);
            int roleid = DefaultDatabaseUtil.getRoleId(dbConnection, roleName);

            if (roleid == -1 || userid == -1) {
                throw new UserStoreException(Messages.getMessage("nullData"));
            }

            PreparedStatement addUserToRoleStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.ADD_USER_ROLE_SQL);
            addUserToRoleStmt.setInt(1, userid);
            addUserToRoleStmt.setInt(2, roleid);
            addUserToRoleStmt.executeUpdate();
            dbConnection.commit();
            addUserToRoleStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorAddingUserToRole"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

    public void removeUserFromRole(String userName, String roleName) throws UserStoreException {
        Connection dbConnection = null;
        try {
            dbConnection = dataSource.getConnection();
            if (dbConnection == null) {
                throw new UserStoreException(Messages.getMessage("null_connection"));
            }
            dbConnection.setAutoCommit(false);
            PreparedStatement removeUserFromRoleStmt = dbConnection
                    .prepareStatement(DefaultRealmConstants.DELETE_USER_ROLE_SQL);
            removeUserFromRoleStmt.setString(1, userName);
            removeUserFromRoleStmt.setString(2, roleName);
            removeUserFromRoleStmt.executeUpdate();
            dbConnection.commit();
            removeUserFromRoleStmt.close();
        } catch (SQLException e) {
            log.debug(e);
            throw new UserStoreException(Messages.getMessage("errorModifyingUserStore"), e);
        } finally {
            try {
                if (dbConnection != null) {
                    dbConnection.close();
                }
            } catch (SQLException e) {
                throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
            }
        }
    }

 
}
