/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.samples.stats;

import org.wso2.registry.jdbc.handlers.UIEnabledHandler;
import org.wso2.registry.jdbc.handlers.RequestContext;
import org.wso2.registry.Resource;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.ResourcePath;
import org.wso2.registry.exceptions.RegistryException;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;

import javax.xml.namespace.QName;
import java.util.Iterator;

/**
 * This Handler generates UIs render ESB statistics resources in many custom views. ESB statistics
 * is a custom resource type, which contains statistics generated from ESB. Their content are in
 * the below XML form:
 *
 *   <esb-statistics>
 *      <server>
 *          <name>AZ gateway</name>
 *          <path>/az/esb</path>
 *      </server>
 *
 *      <message number="12">
 *          <in-time>1240000</in-time>
 *          <out-time>1240002</out-time>
 *          <status>success</status>
 *      </message>
 *  </esb-statistics>
 */
public class ESBStatisticsHandler extends UIEnabledHandler {

    public String[] getBrowseViews() {
        return new String[] { "text:Text view",
                "summary:Summary",
                "details:Detailed view"};
    }

    public String[] getEditViews() {
        return new String[] { "text:Edit as text", "basic:Edit basic information", "details:Edit details" };
    }

    public String[] getNewViews() {
        return new String[] { "plain:Plain view" };
    }

    /**
     * This method generates HTML for three UIs: text, summary and details.
     *
     * @param viewKey UI key of the view.
     * @param requestContext Request details.
     * @return Resource filled with request UI.
     * @throws RegistryException
     */
    public Resource getBrowseView(String viewKey, RequestContext requestContext) throws RegistryException {

        String path = requestContext.getResourcePath().getPath();
        Resource resource = requestContext.getRepository().get(path);

        if (RegistryConstants.TEXT_VIEW_NAME.equals(viewKey)) {

            byte[] contentBytes = (byte[]) resource.getContent();
            String content = new String(contentBytes);
            resource.setContent(content);

        } else if ("summary".equals(viewKey)) {

            OMElement docElement;
            try {
                StAXOMBuilder builder = new StAXOMBuilder(resource.getContentStream());
                docElement = builder.getDocumentElement();
            } catch (Exception ae) {
                throw new RegistryException("Failed to parse the esb statistics file. " +
                        "All statistics should be in XML format.");
            }

            OMElement serverElement = docElement.getFirstChildWithName(new QName("server"));
            OMElement serverNameElement = serverElement.getFirstChildWithName(new QName("name"));
            OMElement serverPathElement = serverElement.getFirstChildWithName(new QName("path"));

            String serverName = serverNameElement.getText();
            String serverPath = serverPathElement.getText();

            String summary = "This entry contains the statistics of the ESB instance <strong>" +
                    serverName + "</strong>.\nMore information and configurations of this " +
                    "ESB instance can be found at " + serverPath + ".";
            resource.setContent(summary);

        } else if ("details".equals(viewKey)) {

            OMElement docElement;
            try {
                StAXOMBuilder builder = new StAXOMBuilder(resource.getContentStream());
                docElement = builder.getDocumentElement();
            } catch (Exception ae) {
                throw new RegistryException("Failed to parse the esb statistics file. " +
                        "All statistics should be in XML format.");
            }

            OMElement serverElement = docElement.getFirstChildWithName(new QName("server"));
            OMElement serverNameElement = serverElement.getFirstChildWithName(new QName("name"));
            OMElement serverPathElement = serverElement.getFirstChildWithName(new QName("path"));

            String serverName = serverNameElement.getText();
            String serverPath = serverPathElement.getText();

            StringBuffer buf = new StringBuffer();
            buf.append("ESB server name: " + serverName);
            buf.append("<br/>ESB server path: " + serverPath);

            buf.append("<br/><table border=1><tr><th>Message number</th>" +
                    "<th>In time</th>" +
                    "<th>Out time</th>" +
                    "<th><Status</th></tr>");

            Iterator messages = docElement.getChildrenWithName(new QName("message"));
            while (messages.hasNext()) {
                OMElement message = (OMElement) messages.next();
                String messageNumber = message.getAttributeValue(new QName("number"));
                String inTime = message.getFirstChildWithName(new QName("in-time")).getText();
                String outTime = message.getFirstChildWithName(new QName("out-time")).getText();
                String status = message.getFirstChildWithName(new QName("status")).getText();

                buf.append("<tr><td>" + messageNumber +
                        "</td><td>" + inTime +
                        "</td><td>" + outTime +
                        "</td><td>" + status + "</td></tr>");
            }
            buf.append("</table><br/>");

            resource.setContent(buf.toString());
        }

        return resource;
    }

    /**
     * This method generates HTML forms for three UIs: text, basic and details.
     *
     * @param editViewKey UI key of the view.
     * @param requestContext Request details.
     * @return Resource filled with requested edit UI content.
     * @throws RegistryException
     */
    public Resource getEditView(String editViewKey, RequestContext requestContext)
            throws RegistryException {

        String path = requestContext.getResourcePath().getPath();
        Resource resource = requestContext.getRepository().get(path);

        if ("text".equals(editViewKey)) {

            byte[] contentBytes = (byte[]) resource.getContent();
            String content = new String(contentBytes);
            resource.setContent(content);

        } else if ("basic".equals(editViewKey)) {

            OMElement docElement;
            try {
                StAXOMBuilder builder = new StAXOMBuilder(resource.getContentStream());
                docElement = builder.getDocumentElement();
            } catch (Exception ae) {
                throw new RegistryException("Failed to parse the esb statistics file. " +
                        "All statistics should be in XML format.");
            }

            OMElement serverElement = docElement.getFirstChildWithName(new QName("server"));
            OMElement serverNameElement = serverElement.getFirstChildWithName(new QName("name"));
            OMElement serverPathElement = serverElement.getFirstChildWithName(new QName("path"));

            String serverName = serverNameElement.getText();
            String serverPath = serverPathElement.getText();

            String view = "<form method='post' action='/wso2registry/custom" + path + "'>" +
                    "<input type='hidden' name='edit-processor' value='application/vnd.esb-stats'/>" +
                    "<input type='hidden' name='view-type' value='edit'/>" +
                    "<input type='hidden' name='view-key' value='basic'/>" +
                    "<h3>Server name: <input type='text' name='server-name' value='" + serverName + "'/></h3>" +
                    "<br/><br/>Server path: <input type='text' name='server-path' value='" + serverPath + "'/>" +
                    "<input type='submit' value='Save ESB information'/>" +
                    "</form>";

            resource.setContent(view);

        } else if ("details".equals(editViewKey)) {

            byte[] contentBytes = (byte[]) resource.getContent();
            String content = new String(contentBytes);

            String view = "<br/><strong>Only XML content are allowed.</strong><br/><form method='post' action='/wso2registry/custom" + path + "'>" +
                    "<input type='hidden' name='edit-processor' value='application/vnd.esb-stats'/>" +
                    "<input type='hidden' name='view-type' value='edit'/>" +
                    "<input type='hidden' name='view-key' value='details'/>" +
                    "<textarea rows='20' cols='70'  name='stats'>" + content + "</textarea>" +
                    "<input type='submit' value='Save ESB information'/>" +
                    "</form>";

            resource.setContent(view);
        }

        return resource;
    }

    public Resource getNewView(String editName, RequestContext requestContext) throws RegistryException {

        ResourcePath resourcePath = requestContext.getResourcePath();
        String parentPath = resourcePath.getParameterValue("parentPath");

        String view = "<strong>New ESB statistics</strong>" +
                "<br/><form method='post' action='/wso2registry/custom" + parentPath + "'>" +
                "<input type='hidden' name='edit-processor' value='application/vnd.esb-stats'/>" +
                "<input type='hidden' name='view-type' value='new'/>" +
                "<input type='hidden' name='view-key' value='plain'/>" +
                "Resource name: <input type='text' name='resourceName'/>" +
                "Content: <textarea rows='20' cols='70'  name='stats'></textarea>" +
                "<input type='submit' value='Save ESB information'/>" +
                "</form>";

        Resource resource = requestContext.getRegistry().newResource();
        resource.setMediaType("application/vnd.esb-stats");
        resource.setContent(view);
        
        return resource;
    }
}
